/* Copyright (C) 2005, 2006, 2007, 2008 by Thomas Hillebrandt and Heiko Will

This file is part of the Micro-mesh SensorWeb Firmware.

Micro-Mesh is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3, or (at your option)
any later version.

Micro-Mesh is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Micro-Mesh; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include <msp430x16x.h>
#include <signal.h>
#include "drivers.h"
#include "boot.h"
#include "type.h"
#include "smallib.h"
#include "driver_system.h"
#include "timer.h"
#include "usertimer.h"
#include "driver_timer.h"
#include "driver_usertimer.h"
#include "config.h"
#include "driver_display.h"
#include "driver_sd.h"
#include "driver_cc1100.h"
#include "driver_sht11.h"
#include "fat.h"
#include "driver_sd.h"
#include "sd.h"
#include "cached_reader.h"
#include "funk.h"
#include "routing.h"
#include "driver_flash.h"
#include "display.h"
#include "net.h"
#include "hardware.h"
#include "driver_adc.h"
#include "msp430/src/std_defines.h"
#include "msp430/src/spi0.h"

/* Global Vars */
#ifdef INCLUDE_FAT16
int log_file_open=0;
file_handle log_file;
#endif

/* Module global Vars*/
#ifdef INCLUDE_SDCARD
	int sdres;
#endif

uint16_t MAXSPEED;
uint16_t MINSPEED;
uint16_t BREAKTIME;
uint16_t XACCTHRESHOLD;

int boot (void)
{
	uint8_t node_id=0;
	#ifdef INCLUDE_BATTERY
		double u;
	#endif

		MAXSPEED = 14;
		MINSPEED = 5;
		BREAKTIME = 600;
		XACCTHRESHOLD = 700;

		/* Init Ports */
		init_ports();

		#ifdef DRIVER
			P4OUT |= 0x02;
		#endif

		/* Init clocks */
		init_clks();

		configure_dev();

		/* Init Console */
		std_devices->stdio_device->driver.init((driver*)std_devices->stdio_device);

		/* Init Timer */
		sl_printf("Init Timer0       ");
		init_timer0();
		sl_printf("[ok]\n\r");

		// Start user timer
		//startUserTimer();

#ifdef DRIVER
		enable_ADC();
		set_3D_sensitivity(ACCELEROMETER_SENSITIVITY_4G);
#endif

		// Enable interrupts on port 1 pin 0 and 1
#ifdef SCHIENE
		P1OUT &= ~(1<<2);
		CLEAR(P1SEL, 0x01);
		CLEAR(P1IES, 0x01);
		SET(P1IE, 0x01);
	    CLEAR(P1IFG, 0x01);

	    CLEAR(P1SEL, 0x02);
	    CLEAR(P1IES, 0x02);
	    SET(P1IE, 0x02);
	    CLEAR(P1IFG, 0x02);

#endif

	#ifdef INCLUDE_SDCARD
		sdres=InitSd();
		if (sdres==0)
		{
			#ifdef INCLUDE_FAT16
			cr_init(sd_read_sector, sd_write_sector);
			fat_init_media(cr_read_sector,cr_write_sector,cr_flush_buffer);
			sl_printf("fatsize: %d\nSectors per Cluster: %d\nbyte per sec: %d\n",vi.fat_size,vi.sec_per_cluster,vi.bytes_per_Sector);
			#endif
		}
	#endif

	#ifdef INCLUDE_DISPLAY
		sl_printf("Init Display      ");
		initDisplay();
		sl_printf("[ok]\n\r");
		glcd_write("Display   [ok]\n");
	#endif

		/* Initialize CC1100 */
		sl_printf("Init CC1100       ");
		if (node_id==0)
		{
			node_id=readNodeId();
		}
		initCC1100(rxBufferAdd,node_id);
		sl_printf("[ok]\n");
		glcd_write("CC1100    [ok]\n");

		/* Initialize all sensors */

		sl_printf("Init Sensors       ");
		int i=0;
		while (std_devices->sensors[i] != NULL)
		{
			std_devices->sensors[i]->base.init(&std_devices->sensors[i]->base);
			i++;
		}
		sl_printf("[ok]\n\r");
		glcd_write("Sensors   [ok]\n");


		/* Initialize Battery */
			#ifdef INCLUDE_BATTERY
				char buffer[10];
				sl_printf("Battery: ");
				u = ReadBatteryVoltage();
				float2string(buffer,u,2,3,1);
				sl_printf(buffer);
				watch_battery();
				sl_printf("V  [ok]\n\r");
				glcd_write("Battery   [ok]\n");

			#endif

			#ifdef INCLUDE_SDCARD
			/* Initialize SDCARD */
				sl_printf("Init SD           ");
				if (sdres==0)
					{
					sl_printf("[ok]\n\r");
					glcd_write("SD        [ok]\n");
					}
				if (sdres==1)
				{
					sl_printf("[Failure 1]");
					glcd_write("SD [FAILURE 1]\n");
				}
				if (sdres==2)
					{
					sl_printf("[Failure 2]");
					glcd_write("SD [FAILURE 2]\n");
					}
				if (sdres==3)
					{
					sl_printf("[no card  ]");
					glcd_write("SD [no card]\n");
					}
			#endif

		/* Initialize random number generator */
		sl_printf("Init Rand       ");
		//sl_srand(getADCSample()); // don't use with enable_ADC
		sl_printf("[ok]\n\r");

		sl_printf("Leds off       ");
		Led1Off();
		Led2Off();
		Led3Off();
		Led4Off();
		sl_printf("[ok]\n\r");
		return 0;
}

int start_routing(void)
{
	/* Initialize the routing layer */
	sl_printf("Init Routing      ");
	/* Initialize the net layer */
	initRouting(net_packet_handler);
	initNet();
	sl_printf("[ok]\n");
	glcd_write("Routing   [ok]\n");
	return 0;
}

int start_logging(void)
{
#ifdef INCLUDE_SDCARD
#ifdef INCLUDE_FAT16
	/* Logging */
	if (sdres==0)
		if (file_openfile(&log_file,"/sensor.log",'r')==0)
		{
			log_file_open=1;
			file_write(&log_file,"<xml>\n<sensor_values>\n",22);
		}
	return 0;
#endif
#endif
	return 1;
}

interrupt (PORT1_VECTOR) lap_counter_isr(void)
{
	if (system_state.POWERDOWN) SPI_INIT; /* Initialize SPI after wakeup */
	/* Check IFG */
	if ((P1IFG & 0x01) != 0)
	{
		CLEAR(P1IFG, 0x01);
		sl_printf("H\n");
	}
	else if ((P1IFG & 0x02) != 0)
	{
		CLEAR(P1IFG, 0x02);
		sl_printf("A\n");
	}
	else
	{
		/* Should not occur - only Port 1 Pin 0 interrupts are enabled */
		CLEAR(P1IFG, 0xFF);	/* Clear all flags */
	}
	if (system_state.POWERDOWN != 0) END_LPM3;
}
