/* Copyright (C) 2005, 2006, 2007, 2008 by Thomas Hillebrandt and Heiko Will

This file is part of the Micro-mesh SensorWeb Firmware.

Micro-Mesh is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3, or (at your option)
any later version.

Micro-Mesh is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Micro-Mesh; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include "smallib.h"
#include "funk.h"
#include "type.h"
#include "defaultSettings.h"
#include "helpers.h"
#include "timer.h"
#include "driver_cc1100.h"
#include "driver_system.h"
#include "driver_timer.h"
#include "cc1100_defines.h"
#include "user_apps.h"
#include "boot.h"


#define MAX_UID				(0xFF) // Maximum UID of a node is 255
#define MIN_UID				(0x01) // Minimum UID of a node is 1
#define MAX_WINDOW_SIZE		(0x14) // Maximum window size of backoff algorithm
#define MAX_SEQ_BUFFER_SIZE   (20) // Maximum size of the sequence number buffer

volatile uint8_t uid = 0;						// Initial sensornode address is 0
volatile uint8_t radioState = RADIO_UNKNOWN;	// Initial radio state

void cc1100_carrier_sense(void);
void cc1100_packet_in(void);

uint8_t paTableIndex = PATABLE;	// Current PATABLE Index
uint8_t paTable[] = {0x00,		// -52 dBm
				   0x03,		// -30 dBm
				   0x0D,		// -20 dBm
				   0x1C,		// -15 dBm
				   0x34,		// -10 dBm
				   0x57,		// - 5 dBm
				   0x3F,		// - 1 dBm
				   0x8E,		//	 0 dBm
				   0x85,		// + 5 dBm
				   0xCC,		// + 7 dBm
				   0xC6, 		// + 9 dBm
				   0xC3  		// +10 dBm
				  };

static uint8_t rxBuffer[62];				// RX Buffer
static packet_layer0_t txBuffer;		// TX Buffer

static volatile int worTimerId;	// Timer ID for WOR
static volatile routing_callback_t routingCallback;

//------------------------- Sequence number buffer ------------------------------
typedef struct
{
	uint8_t source;
	uint8_t identification;
	uint32_t mTicks;
} seqBufferEntry;

static seqBufferEntry seqBuffer[MAX_SEQ_BUFFER_SIZE];	// Sequence number buffer for this layer
static uint8_t seqBufferPos = 0;							// Next pos to enter a new value into ::seqBuffer

static volatile uint16_t lastSeqNum = 0;					// Last sequence number this node has seen
														// (phySrc + flags.identification) - for speedup in ISR.

//-------------------------------------------------------------------------------------------------------
// Sequence number buffer management
//-------------------------------------------------------------------------------------------------------

bool containsLayer0SeqEntry(uint8_t src, uint8_t id)
{
	uint8_t i;
	uint32_t now = *TICKS();
	for (i = 0; i < MAX_SEQ_BUFFER_SIZE; i++)
	{
		if ((seqBuffer[i].source == src) && (seqBuffer[i].identification == id))
		{
			// Check if timestamp is ok
			if ((now - seqBuffer[i].mTicks) <= SEQ_VALID_TIME)
			{
				return true;
			}
			else
			{
				seqBuffer[i].source = 0; // Reset
			}
		}
	}
	return false;
}

void addLayer0SeqEntry(uint8_t src, uint8_t id)
{
	// Remove all entries with given source to avoid short time overflow
	// of one bit counter (of the source node). So a valid packet would get lost
	uint8_t i;
	for (i = 0; i < MAX_SEQ_BUFFER_SIZE; i++)
	{
		if (seqBuffer[i].source == src)
		{
			seqBuffer[i].source = 0; // Reset
		}
	}

	// Add new entry
	seqBuffer[seqBufferPos].source = src;
	seqBuffer[seqBufferPos].identification = id;
	seqBuffer[seqBufferPos].mTicks = *TICKS();

	// Store 16 bit sequence number of layer 0 for speedup
	lastSeqNum = src;
	lastSeqNum <<= 8;
	lastSeqNum += id;

	seqBufferPos++;
	if (seqBufferPos == MAX_SEQ_BUFFER_SIZE) seqBufferPos = 0;
}

//-------------------------------------------------------------------------------------------------------
//  void spiWriteReg(uint8_t addr, uint8_t value)
//
//  DESCRIPTION:
//      Function for writing to a single CC1100 register.
//
//  ARGUMENTS:
//      uint8_t addr
//          Address of a specific CC1100 register to accessed.
//      uint8_t value
//          Value to be written to the specified CC1100 register.
//-------------------------------------------------------------------------------------------------------
void spiWriteReg(uint8_t addr, uint8_t value)
{
	drivercc1100_spiinittrx();
	drivercc1100_trxspi(addr);
	drivercc1100_trxspi(value);
	CC1100_CS_HIGH;
}

//-------------------------------------------------------------------------------------------------------
//  uint8_t spiReadReg(uint8_t addr)
//
//  DESCRIPTION:
//      This function gets the value of a single specified CC1100 register.
//
//  ARGUMENTS:
//      uint8_t addr
//          Address of the CC1100 register to be accessed.
//
//  RETURN VALUE:
//      uint8
//          Value of the accessed CC1100 register.
//-------------------------------------------------------------------------------------------------------
uint8_t spiReadReg(uint8_t addr)
{
	uint8_t x;
	drivercc1100_spiinittrx();
	drivercc1100_trxspi(addr | CC1100_READ_SINGLE);
	x = drivercc1100_trxspi(NOBYTE);
	CC1100_CS_HIGH;
	return x;
}

//-------------------------------------------------------------------------------------------------------
//  void spiStrobe(uint8_t strobe)
//
//  DESCRIPTION:
//      Function for writing a strobe command to the CC1100
//
//  ARGUMENTS:
//      uint8_t strobe
//          Strobe command
//-------------------------------------------------------------------------------------------------------
void spiStrobe(uint8_t strobe)
{
	drivercc1100_spiinittrx();
	drivercc1100_trxspi(strobe);
	CC1100_CS_HIGH;
}

//-------------------------------------------------------------------------------------------------------
//  uint8_t spiReadStatus(uint8_t addr)
//
//  DESCRIPTION:
//      This function reads a CC1100 status register.
//
//  ARGUMENTS:
//      uint8_t addr
//          Address of the CC1100 status register to be accessed.
//
//  RETURN VALUE:
//      uint8
//          Value of the accessed CC1100 status register.
//-------------------------------------------------------------------------------------------------------
uint8_t spiReadStatus(uint8_t addr)
{
	uint8_t x;
	drivercc1100_spiinittrx();
	drivercc1100_trxspi(addr | CC1100_READ_BURST);
	x = drivercc1100_trxspi(NOBYTE);
	CC1100_CS_HIGH;
	return x;
}

//-------------------------------------------------------------------------------------------------------
//  uint8_t spiGetRxStatus(void)
//
//  DESCRIPTION:
//      This function transmits a No Operation Strobe (SNOP) with the read bit set to get the status of
//      the radio.
//
//      Status byte:
//
//      --------------------------------------------------------------------------------
//      |          |            |                                                      |
//      | CHIP_RDY | STATE[2:0] | FIFO_BYTES_AVAILABLE (available bytes in the RX FIFO)|
//      |          |            |                                                      |
//      --------------------------------------------------------------------------------
//
//      STATE[2:0]:
//
//      Value | State
//      --------------------------
//      000   | Idle
//      001   | RX
//      010   | TX
//      011   | FSTXON
//      100   | CALIBRATE
//      101   | SETTLING
//      110   | RX_FIFO_OVERFLOW
//      111   | TX_FIFO_UNDERFLOW
//-------------------------------------------------------------------------------------------------------
uint8_t spiGetRxStatus(void)
{
	uint8_t x;
	drivercc1100_spiinittrx();
	// Strobe SNOP with the read byte set to get info on
	// current state and number of bits in the RX FIFO
	x = drivercc1100_trxspi(CC1100_SNOP | CC1100_READ_SINGLE);
	CC1100_CS_HIGH;
	return x;
}

//-------------------------------------------------------------------------------------------------------
//  uint8_t spiGetTxStatus(void)
//
//  DESCRIPTION:
//      This function transmits a No Operation Strobe (SNOP) to get the status of the radio.
//
//      Status byte:
//
//      ---------------------------------------------------------------------------
//      |          |            |                                                 |
//      | CHIP_RDY | STATE[2:0] | FIFO_BYTES_AVAILABLE (free bytes in the TX FIFO)|
//      |          |            |                                                 |
//      ---------------------------------------------------------------------------
//
//      STATE[2:0]:
//
//      Value | State
//      --------------------------
//      000   | Idle
//      001   | RX
//      010   | TX
//      011   | FSTXON
//      100   | CALIBRATE
//      101   | SETTLING
//      110   | RX_FIFO_OVERFLOW
//      111   | TX_FIFO_UNDERFLOW
//-------------------------------------------------------------------------------------------------------
uint8_t spiGetTxStatus(void)
{
	uint8_t x;
	drivercc1100_spiinittrx();
	// Strobe SNOP to get info on current state and number
	// of bits in the TX FIFO
	x = drivercc1100_trxspi(CC1100_SNOP);
	CC1100_CS_HIGH;
	return x;
}

//-------------------------------------------------------------------------------------------------------
//  void setWorMode(void)
//
//  DESCRIPTION:
//      Function for setting the chip into WOR (Wake-On-Radio) mode.
//-------------------------------------------------------------------------------------------------------
void setWorMode(void)
{
	// If chip ist already in WOR mode do nothing
	if (radioState == RADIO_WOR) return;

	// Make sure CC1100 is in IDLE state
	spiStrobe(CC1100_SIDLE);

	// Enable automatic initial calibration of RCosc.
	// Set T_event1 ~ 1.4 ms, enough for XOSC stabilize and FS calibration before RX.
    // Enable RC oscillator before starting with WOR (or else it will not wake up).
	// Not using AUTO_SYNC function.
	spiWriteReg(CC1100_WORCTRL, WORCTRL);

	// Set Event0 timeout (RX polling interval)
	spiWriteReg(CC1100_WOREVT1, WOREVT1);
    spiWriteReg(CC1100_WOREVT0, WOREVT0);

	// Set RX time in WOR mode
	spiWriteReg(CC1100_MCSM2, WOR_RX_TIME);

	// Enable automatic FS calibration when going from IDLE to RX/TX/FSTXON (in between EVENT0 and EVENT1)
    spiWriteReg(CC1100_MCSM0, 0x18);

	// Put the radio to SLEEP by starting Wake-on-Radio.
	rflags.RX_REQ = false;
    spiStrobe(CC1100_SWORRST);	// Resets the real time clock
    spiStrobe(CC1100_SWOR);		// Starts Wake-on-Radio
	radioState = RADIO_WOR;
}

//-------------------------------------------------------------------------------------------------------
//  void switchToWOR(void)
//
//  DESCRIPTION:
//		Sets the radio to WOR (Wake-On-Radio) mode after a packet has been
//		received (interrupt is handled).
//
//-------------------------------------------------------------------------------------------------------
void switchToWOR()
{
	if (CC1100_GDO2)						// Any incoming packet?
	{
		return;								// then don't go to WOR now
	}

	if (!rflags.MAN_WOR)
	{
		rflags.MAN_WOR = true;
		radioState = RADIO_WOR;
		spiStrobe(CC1100_SIDLE);
		spiStrobe(CC1100_SPWD);
		worTimerId = register_timer(switchToWOR, irand(WOR_MAN_SLEEP_TIME) + 1);
	}
	else
	{
		rflags.MAN_WOR = false;
		WOR_WAKEUP();
		spiStrobe(CC1100_SRX);
		wait_ticks(1);
		radioState = RADIO_RX;
		uint32_t tmp = *TICKS() + WOR_MAN_RX_TIME;
		while (*TICKS() < tmp)
		{
			if (rflags.RX_REQ)
			{
				return;
			}
		}
		spiStrobe(CC1100_SIDLE);				// Put CC1100 to IDLE
		radioState = RADIO_IDLE;				// Radio state now IDLE
		if (rflags.RX_REQ) return;
		spiWriteReg(CC1100_MCSM2, WOR_RX_TIME);	// Configure RX_TIME (for use in WOR)
		spiWriteReg(CC1100_MCSM0, 0x18);		// Turn on FS-Autocal
		spiStrobe(CC1100_SWOR);					// Put radio back to sleep/WOR (must be in IDLE when this is done)
		radioState = RADIO_WOR;					// Radio state now WOR
	}
}

//-------------------------------------------------------------------------------------------------------
//  void WOR_WAKEUP(void)
//
//  DESCRIPTION:
//      This function must be called for CC1100 WOR-Wakeup. Only after returning from this function valid
//      operation with the CC1100 is possible.
//-------------------------------------------------------------------------------------------------------
void WOR_WAKEUP(void)
{
	if (radioState != RADIO_WOR) return;

	// Wake up the chip from WOR/sleep
	drivercc1100_spiinittrx();
	wait_ticks(1);
	CC1100_CS_HIGH;

	radioState = RADIO_IDLE;

	// XOSC startup + FS calibration (300 + 809 us  ~ 1.38 ms)
	wait_ticks(FS_CAL_TIME);
}

//-------------------------------------------------------------------------------------------------------
// CC1100 reset functionality
//-------------------------------------------------------------------------------------------------------

void reset(void)
{
	WOR_WAKEUP();
	drivercc1100_spiinittrx();
	drivercc1100_trxspi(CC1100_SRES);
	drivercc1100_spiinittrx();
	CC1100_CS_HIGH;
	wait_ticks(1);
}

//-------------------------------------------------------------------------------------------------------
//	void powerUpReset(void)
//
//	DESCRIPTION:
//		Reset the CC1100 after power_on and wait for it to be ready.
//
//                 min 40 us
//                 <------------------>
//  CSn     |--|  |--------------------|            |-----
//          |  |  |                    |            |
//              --                      ------------
//
//  MOSI                                    |----|
//          -----------------------------|  |    |
//                                        --      ---------
//               Unknown / don't care
//                                       SRES     done
//-------------------------------------------------------------------------------------------------------
void powerUpReset(void)
{
	CC1100_CS_HIGH;
	CC1100_CS_LOW;
	CC1100_CS_HIGH;
	wait_ticks(RESET_WAIT_TIME);
	reset();
}

//-------------------------------------------------------------------------------------------------------
// CC1100 initialisation and configuration code
//-------------------------------------------------------------------------------------------------------

void initCC1100(routing_callback_t callback,uint8_t id)
{
	drivercc1100_load(cc1100_carrier_sense,cc1100_packet_in);
	powerUpReset();
	// Write configuration to config registers
	drivercc1100_spiwriteburstreg(0x00, conf, sizeof(conf));
	// Write PATABLE (power settings)
	spiWriteReg(CC1100_PATABLE, paTable[paTableIndex]);

	// Initialize Radio Flags
	rflags.RSSI = 0x00;
	rflags.LL_ACK = false;
	rflags.CAA = false;
	rflags.CRC = false;
	rflags.SEQ = false;
	rflags.MAN_WOR = false;
	worTimerId = -1;
	routingCallback = callback;

	uint8_t i;
	for (i = 0; i < MAX_SEQ_BUFFER_SIZE; i++)
	{
		seqBuffer[i].source = 0;
	}

	setUid(id);

	drivercc1100_initinterrupts();

	// Set chip to RX mode
	spiWriteReg(CC1100_MCSM2, 0x07);
	spiStrobe(CC1100_SRX);
}

void setUid(uint8_t id)
{
	if (id < MIN_UID) return;
	// Save old radio state
	uint8_t oldState = radioState;
	// Wake up from WOR (if in WOR, else no effect)
	WOR_WAKEUP();
	// Save new UID here and in CC1100 register
	uid = id;
	spiWriteReg(CC1100_ADDR, id);
	// Have to put radio back to WOR if old radio state
	// was WOR, otherwise no action is neccessary
	if (oldState == RADIO_WOR)
	{
		spiStrobe(CC1100_SWOR);	// Put radio back to sleep/WOR (must be in IDLE when this is done)
		radioState = RADIO_WOR;	// Radio state now WOR
	}
}

bool setOutputPower(uint8_t paIdx) {
	if (paIdx < 12) {
		// Save old radio state
		uint8_t oldState = radioState;
		// Wake up from WOR (if in WOR, else no effect)
		WOR_WAKEUP();
		// Set new PATABLE Index
		paTableIndex = paIdx;
		// And write to CC1100 PATABLE register
		spiWriteReg(CC1100_PATABLE, paTable[paTableIndex]);
		// Have to put radio back to WOR if old radio state
		// was WOR, otherwise no action is neccessary
		if (oldState == RADIO_WOR) {
			spiStrobe(CC1100_SWOR);	// Put radio back to sleep/WOR (must be in IDLE when this is done)
			radioState = RADIO_WOR;	// Radio state now WOR
		}
		return true;
	}
	return false;
}

signed short getLastRSSI(void) {
	signed short ret = rflags.RSSI;
	if (rflags.RSSI >= 128)
		ret -= 256;
	ret /= 2;
	ret -= 78;
	return ret;
}

//-------------------------------------------------------------------------------------------------------
// CC1100 send and receive functionality
//-------------------------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------------------------
//  void sendPacket(uint8_t *txBuffer, uint8_t size)
//
//  DESCRIPTION:
//      This function can be used to transmit a packet with packet length up to 62 bytes.
//      To use this function, GD02 must be configured to be asserted when sync word is sent and
//      de-asserted at the end of the packet => spiWriteReg(CC1100_IOCFG0, 0x06);
//      The function implements polling of GDO2. First it waits for GD02 to be set and then it waits
//      for it to be cleared.
//
//  ARGUMENTS:
//      uint8_t *txBuffer
//          Pointer to a buffer containing the data that are going to be transmitted.
//
//      uint8_t size
//          The size of the txBuffer (number of bytes to be read).
//-------------------------------------------------------------------------------------------------------
void sendPacket(uint8_t *txBuffer, uint8_t size)
{
	drivercc1100_beforesend();

	// The number of bytes to be transmitted must be smaller
	// or equal to PACKET_LENGTH (62 bytes). So the receiver
	// can put the whole packet in its RX-FIFO (with appended
	// packet status bytes).
	if (size > PACKET_LENGTH) return;

	// But CC1100 in IDLE mode to flush the FIFO
	spiStrobe(CC1100_SIDLE);

    // Flush TX FIFO to be sure it is empty
    spiStrobe(CC1100_SFTX);
	// Write packet into TX FIFO

    drivercc1100_spiwriteburstreg(CC1100_TXFIFO, txBuffer, size);

  	// Switch to TX mode
	spiStrobe(CC1100_STX);

	long c = 0;
	while (!CC1100_GDO2)
	{
		if (c++ == 1000000)
			alarm();
	}

	// Wait for GDO2 to be cleared -> end of packet
	c = 0;
	while (CC1100_GDO2)
	{
		if (c++ == 1000000)
			alarm();
	}

	// Enable interrupts for GDO2
	drivercc1100_aftersend();
}

//-------------------------------------------------------------------------------------------------------
//	bool sendLinkLevelAck(uint8_t dest)
//
//	DESCRIPTION:
//		Send a link level acknowledgement to the specified sensornode.
//
//		---------------------------------------------------------------------
//		|        |         |                          |          |          |
//		| Length | Address | PhySrc (Physical Source) |  Flags   |   Data   |
//		|   = 3  | = dest  |         = uid            | = L1_ACK |          |
//		---------------------------------------------------------------------
//		  1 byte   1 byte             1 byte            1 byte     0 bytes
//
//		=> Packet size = 14/18 bytes (4/8 PRE + 4 SYNC + 4 L&A&PS&F + 2 CRC) (depends on datarate).
//      => Send-Time@19.2 kbps = 14 * 8 bit / 19.2 kbps = 5,83 ms
//		=> Send-Time@100  kbps = 14 * 8 bit / 100  kbps = 1,12 ms
//		=> Send-Time@400  kbps = 18 * 8 bit / 400  kbps = 0,36 ms
//
//  ARGUMENTS:
//      uint8_t dest
//			The destination address of the link level acknowledgement packet.
//
//-------------------------------------------------------------------------------------------------------
void sendLinkLevelAck(uint8_t dest)
{
	uint8_t oldState = radioState;		// Save old state
	radioState = RADIO_SEND_ACK;		// Set state to "Sending ACK"
	spiWriteReg(CC1100_MCSM0, 0x08);	// Turn off FS-Autocal
	spiWriteReg(CC1100_MCSM1, 0x00);	// TX_OFFMODE = IDLE
	packet_layer0_t ack;				// Local packet, don't overwrite
	ack.length = 3;						// possible packet in txBuffer!
	ack.address = dest;
	ack.phySrc = uid;
	ack.flags = (L1_ACK << 1) & 0x0E;
	sendPacket((uint8_t*)&ack, 4);		// IDLE -> TX (88.4 us)
	spiWriteReg(CC1100_MCSM0, 0x18);	// Turn on FS-Autocal
	spiWriteReg(CC1100_MCSM1, 0x03);	// TX_OFFMODE = RX
	radioState = oldState;				// Restore state
}

//-------------------------------------------------------------------------------------------------------
//	bool sendBurst(packet_layer0_t* packet)
//
//	DESCRIPTION:
//		Sends a packet in burst mode. After sending a single packet, the radio is listening
//		for a possible ACK (link level). If an ACK is received the burst is interrupted.
//
//  ARGUMENTS:
//      packet_layer0_t* packet
//			Pointer to the outgoing packet.
//		uint8_t retries
//			The number of retries left.
//
//  RETURN VALUE:
//		bool
//			true:	Always _true_ if the address field of the packet was a broadcast address;
//					Also _true_ if an ACK has been received.
//			false:  If packet address field was no broadcast address and no ACK was received.
//-------------------------------------------------------------------------------------------------------
bool sendBurst(packet_layer0_t* packet, uint8_t retries)
{
	radioState = RADIO_SEND_BURST;
	rflags.LL_ACK = false;
	uint8_t i = BURST_COUNT;

	while (i-- > 0)
	{
		sendPacket((uint8_t*)packet, PACKET_LENGTH);	// RX -> TX (9.6 us)
		wait_ticks(TIMEOUT_SEND);			// After sending the packet the CC1100 goes automatically
											// into RX mode (21.5 us) (listening for an ACK).

		if (rflags.LL_ACK) break;
	}

	// Packet was unicast and no link level ACK -> do retry if allowed
	if (packet->address != ADDR_BROADCAST && !rflags.LL_ACK && retries > 0)
	{
		return sendBurst(packet, retries - 1);
	}

	switchToWOR();

	lastSeqNum = 0; // Burst from any other node is definitly over

	if (packet->address != ADDR_BROADCAST && !rflags.LL_ACK)
	{
		return false;
	}

	return true;
}

//-------------------------------------------------------------------------------------------------------
//	bool send(packet_layer0_t* packet)
//
//	DESCRIPTION:
//		Sends a packet on the MAC layer. The sending algorithm is similar to IEEE-802.11 (CSMA/CA).
//
//  ARGUMENTS:
//      packet_layer0_t* packet
//			Pointer to the outgoing packet.
//
//  RETURN VALUE:
//		bool
//			true:	If packet was send successfully.
//			false:  If packet could not be send or no ACK was received.
//-------------------------------------------------------------------------------------------------------
bool send(packet_layer0_t* packet)
{
	WOR_WAKEUP();						// Wake CC1100 up from Wake-On-Radio mode
	if (radioState == RADIO_RX)			// If radio in RX mode
	{
		spiStrobe(CC1100_SIDLE);		// Go back to IDLE for calibration
	}
	spiWriteReg(CC1100_MCSM0, 0x08);	// Turn off FS-Autocal
	spiStrobe(CC1100_SCAL);				// Calibrate manually (721 us)
	wait_ticks(FS_CAL_MAN_TIME);		// Wait for calibration to finish before packet burst can start
	radioState = RADIO_AIR_FREE_WAITING;// Set status "waiting for air free"
	spiWriteReg(CC1100_MCSM2, 0x07);	// Configure RX_TIME = Until end of packet (no timeout)
	spiStrobe(CC1100_SRX);				// Switch to RX (88.4 us) (Carrier Sense)
	wait_ticks(RX_2_CS_TIME);			// Wait until CC1100 is in RX + carrier sense ready (GDO0 ready for readout -> datarate dependant!!!)

	uint16_t windowSize = 2;	// Start with window size of 4 (max: MAX_WINDOW_SIZE)
	uint16_t backoff = 0;		// Backoff between 1 and windowSize
	uint32_t total;			// Holds the total wait time before send try

	Led2On();			// Turn red led on to indicate wait state

	window:
		if (backoff != 0) goto cycle;				// If backoff was 0
		windowSize *= 2;							// ...increase window size
		if (windowSize > MAX_WINDOW_SIZE)
		{
			windowSize = MAX_WINDOW_SIZE;			// This is the maximum size allowed
		}
		backoff = irand(windowSize) + (uint16_t) 1;	// ...and choose new backoff
	cycle:
		while (CC1100_GDO0);						// GDO0 reflects CS (high: air not free, low: air free)
		rflags.CAA = true;							// Air is free now
		drivercc1100_enablecsdetect();
		if (backoff > 0) backoff--;					// Decrement backoff counter

		total = SLOTTIME;							// Calculate total wait time
		total *= (uint32_t)backoff;					// Slot vector set
		total += *TICKS();							// Add current time
		total += DIFS;								// ...and standard DIFS wait time
		total += 10000;								// Offset to avoid underflow in while loop

		while ((total - *TICKS()) > 10000)			// Wait until timeout is finished
		{
			if (!rflags.CAA)						// Is the air still free?
			{
				goto window;						// No. Go back to new wait period.
			}
		}

		Led2Off();									// Wait state is over now
		drivercc1100_disablecsdetect();

		return sendBurst(packet, BURST_RETRY_COUNT);// Send raw burst

	// Packet could not be send, return false
	return false;
}

bool csend(uint8_t addr, uint8_t protocol, uint8_t *data, uint8_t length)
{
	uint8_t i;
	if (addr == uid) return false;		// Loopback not supported

	if (length > MAX_DATA_LENGTH)		// Check length
	{
		length = MAX_DATA_LENGTH;		// Correct length
	}

	for (i = 0; i < MAX_DATA_LENGTH; i++)
	{
		txBuffer.data[i] = 0;			// Clean data
	}

	txBuffer.length = 3 + length;				// 3 Bytes (A&PS&F) + data length
	txBuffer.address = addr;					// Copy destination address
	txBuffer.flags = 0x00;						// Set clean state
	txBuffer.flags = W_FLAGS_PROTOCOL(protocol);// Copy protocol identifier

	// Set the physical sender of this packet to the uid of this sensornode,
	// so LL-ACK can be send back.
	txBuffer.phySrc = uid;

	// Set identification number of packet
	txBuffer.flags |= rflags.SEQ;	// Set flags.identification (bit 0)
	rflags.SEQ = !rflags.SEQ;		// Toggle value of layer 0 sequence number bit

	for (i = 0; i < length; i++)
	{
		txBuffer.data[i] = data[i];			// Copy data
	}
	return send(&txBuffer);					// Send the packet
}

//-------------------------------------------------------------------------------------------------------
//  bool receivePacketVariable(uint8_t *rxBuffer, uint8_t length)
//
//  DESCRIPTION:
//      This function can be used to receive a packet of variable packet length (first byte in the packet
//      must be the length byte). The packet length should not exceed the RX FIFO size.
//      To use this function, APPEND_STATUS in the PKTCTRL1 register must be enabled.
//      The function must be called in an ISR that is triggered when a valid packet is received (GDO pins).
//      The RXBYTES register is read to make sure that there are bytes in the FIFO. This is because the
//      GDO signal will indicate sync received even if the FIFO is flushed due to address filtering, CRC
//		filtering, or packet length filtering.
//
//  ARGUMENTS:
//      uint8_t *rxBuffer
//          Pointer to the buffer where the incoming data should be stored.
//      uint8_t length
//          The size of the buffer where the incoming data should be stored.
//
//  RETURN VALUE:
//      bool
//          true:   CRC OK
//          false:  CRC NOT OK or storage buffer too short (or no packet was put in the RX FIFO due to filtering)
//-------------------------------------------------------------------------------------------------------
bool receivePacketVariable(uint8_t *rxBuffer, uint8_t length)
{
	uint8_t status[2];
	uint8_t packetLength = 0;

	// Any bytes available in RX FIFO?
	if ((spiReadStatus(CC1100_RXBYTES) & BYTES_IN_RXFIFO))
	{
		// Read length byte (first byte in RX FIFO)
        packetLength = spiReadReg(CC1100_RXFIFO);

		// Read data from RX FIFO and store in rxBuffer
        if (packetLength <= length)
		{
			// Put length byte at first position in RX Buffer
			rxBuffer[0] = packetLength;

			// Read the rest of the packet
            drivercc1100_spireadburstreg(CC1100_RXFIFO, rxBuffer+1, packetLength);

            // Read the 2 appended status bytes (status[0] = RSSI, status[1] = LQI)
            drivercc1100_spireadburstreg(CC1100_RXFIFO, status, 2);

			// Store RSSI value of packet
			rflags.RSSI = status[I_RSSI];

			// MSB of LQI is the CRC_OK bit
			rflags.CRC = (status[I_LQI] & CRC_OK) >> 7;

			return rflags.CRC;
        }
		else
		{
			// Make sure that the radio is in IDLE state before flushing the FIFO
            // (Unless RXOFF_MODE has been changed, the radio should be in IDLE state at this point)
            spiStrobe(CC1100_SIDLE);

            // Flush RX FIFO
            spiStrobe(CC1100_SFRX);
            return false;
        }
	}
	else
	{
		return false;
	}
}

//-------------------------------------------------------------------------------------------------------
//  bool receivePacketFixed(uint8_t *rxBuffer, uint8_t length)
//
//  DESCRIPTION:
//      This function can be used to receive a packet of fixed packet length (length is defined in
//		the PKTLEN register). The packet length should not exceed the RX FIFO size.
//      To use this function, APPEND_STATUS in the PKTCTRL1 register must be enabled.
//      The function must be called in an ISR that is triggered when a valid packet is received (GDO pins).
//      The RXBYTES register is read to make sure that there are bytes in the FIFO. This is because the
//      GDO signal will indicate sync received even if the FIFO is flushed due to address filtering, CRC
//		filtering, or packet length filtering.
//
//  ARGUMENTS:
//      uint8_t *rxBuffer
//          Pointer to the buffer where the incoming data should be stored.
//      uint8_t length
//          The size of the buffer where the incoming data should be stored.
//
//  RETURN VALUE:
//      bool
//          true:   CRC OK
//          false:  CRC NOT OK or storage buffer too short (or no packet was put in the RX FIFO due to filtering)
//-------------------------------------------------------------------------------------------------------
bool receivePacketFixed(uint8_t *rxBuffer, uint8_t length)
{
	uint8_t status[2];
	uint8_t packetLength = 0;

	// Read fixed packet length out of PKTLEN register
	packetLength = spiReadReg(CC1100_PKTLEN);

	// Any bytes available in RX FIFO?
	if ((spiReadStatus(CC1100_RXBYTES) & BYTES_IN_RXFIFO))
	{
		// Read data from RX FIFO and store in rxBuffer
        if (packetLength <= length)
		{
        	drivercc1100_spireadburstreg(CC1100_RXFIFO, rxBuffer, packetLength);

            // Read the 2 appended status bytes (status[0] = RSSI, status[1] = LQI)
            drivercc1100_spireadburstreg(CC1100_RXFIFO, status, 2);

			// Store RSSI value of packet
			rflags.RSSI = status[I_RSSI];

            // MSB of LQI is the CRC_OK bit
			rflags.CRC = (status[I_LQI] & CRC_OK) >> 7;

			return rflags.CRC;
        }
		else
		{
			// Make sure that the radio is in IDLE state before flushing the FIFO
            // (Unless RXOFF_MODE has been changed, the radio should be in IDLE state at this point)
            spiStrobe(CC1100_SIDLE);

            // Flush RX FIFO
            spiStrobe(CC1100_SFRX);
            return false;
        }
	}
	else
	{
		return false;
	}
}

//-------------------------------------------------------------------------------------------------------
//  bool receivePacket(uint8_t *rxBuffer, uint8_t length)
//
//  DESCRIPTION:
//      This function can be used to receive a packet of fixed or of variable packet length. The right
//		option is choosen by looking in the CC1100 config register. The packet length should not exceed
//		the RX FIFO size.
//      To use this function, APPEND_STATUS in the PKTCTRL1 register must be enabled.
//      The function must be called in an ISR that is triggered when a valid packet is received (GDO pins).
//      The RXBYTES register is read to make sure that there are bytes in the FIFO. This is because the
//      GDO signal will indicate sync received even if the FIFO is flushed due to address filtering, CRC
//		filtering, or packet length filtering.
//
//  ARGUMENTS:
//      uint8_t *rxBuffer
//          Pointer to the buffer where the incoming data should be stored.
//      uint8_t length
//          The size of the buffer where the incoming data should be stored.
//
//  RETURN VALUE:
//      bool
//          true:   CRC OK
//          false:  CRC NOT OK or storage buffer too short (or no packet was put in the RX FIFO due to filtering)
//-------------------------------------------------------------------------------------------------------
bool receivePacket(uint8_t *rxBuffer, uint8_t length)
{
	uint8_t pktlencfg = spiReadReg(CC1100_PKTCTRL0) & PKT_LENGTH_CONFIG;
	if (pktlencfg == FIXED_PKTLEN)
	{
		return receivePacketFixed(rxBuffer, length);
	}
	else if (pktlencfg == VARIABLE_PKTLEN)
	{
		return receivePacketVariable(rxBuffer, length);
	}
	return false;
}

//-------------------------------------------------------------------------------------------------------
// CC1100 receive interrupt
//-------------------------------------------------------------------------------------------------------

void cc1100_carrier_sense(void)
{
	rflags.CAA = false;		// Air was not free, so clear CCA flag
}

static volatile int hysteresis = 0;
static volatile int started = 0;

void set_speed(int speed) {
	if (started == 0) return;
	P1OUT |= (1<<2);
	register int i = 0;
	P1OUT &= ~(1<<6);
	for (i=0; i < 100; i++) {
		P1OUT &= ~(1<<7);
		_NOP(); _NOP(); _NOP(); _NOP(); _NOP();
		_NOP(); _NOP(); _NOP(); _NOP(); _NOP();
		P1OUT |= (1<<7);
		_NOP(); _NOP(); _NOP(); _NOP(); _NOP();
		_NOP(); _NOP(); _NOP(); _NOP(); _NOP();
	}
	P1OUT |= (1<<6);
	_NOP(); _NOP(); _NOP(); _NOP(); _NOP();
	_NOP(); _NOP(); _NOP(); _NOP(); _NOP();
	for (i=0; i < speed; i++) {
		P1OUT &= ~(1<<7);
		_NOP(); _NOP(); _NOP(); _NOP(); _NOP();
		_NOP(); _NOP(); _NOP(); _NOP(); _NOP();
		P1OUT |= (1<<7);
		_NOP(); _NOP(); _NOP(); _NOP(); _NOP();
		_NOP(); _NOP(); _NOP(); _NOP(); _NOP();
	}
}

void start_car(void)
{
	started = 1;
	set_speed(MAXSPEED);
	sl_printf("T\n");
}

void stop_car(void)
{
	started = 0;
	if (hysteresis != 0) delete_timer(hysteresis);
	P1OUT &= ~(1<<2);
}

static void throttle(void)
{
	set_speed(MAXSPEED);
	hysteresis = 0;
	sl_printf("T\n");
}

void cc1100_packet_in(void)
{

	// Possible packet received, RX -> IDLE (0.1 us)
	rflags.CAA = false;
	rflags.MAN_WOR = false;

	// Transfer packet into buffer
	bool res = receivePacket(rxBuffer, sizeof(rxBuffer));

	if (res)
	{
		int x = ((uint16_t*)rxBuffer)[1];
		x -= 2048;
		x = sl_abs(x);
		if (x > XACCTHRESHOLD && hysteresis == 0)
		{
			set_speed(MINSPEED);
			hysteresis = register_timer(throttle, BREAKTIME);
			sl_printf("B\n");
		}
#ifdef GUI
		sl_printf("%u,%u,%u,%u,%u,%u,%u\r\n", ((uint16_t*)rxBuffer)[5],
				((uint16_t*)rxBuffer)[6], ((uint16_t*)rxBuffer)[7],
				((uint16_t*)rxBuffer)[4], ((uint16_t*)rxBuffer)[1],
				((uint16_t*)rxBuffer)[2],((uint16_t*)rxBuffer)[3]);
#endif

	}
	else
	{
		lastSeqNum = 0;				// Reset for correct burst detection
		spiStrobe(CC1100_SIDLE);	// Switch to IDLE (should already be)...
		spiStrobe(CC1100_SFRX);		// ...for flushing the RX FIFO
	}

	//spiWriteReg(CC1100_MCSM2, 0x07);
	spiStrobe(CC1100_SRX);
}

void handleRadioEvents(void)
{
	if (rflags.RX_REQ)
	{
		rflags.RX_REQ = false;
		spiWriteReg(CC1100_MCSM0, 0x08);	// Turn off FS-Autocal
		spiWriteReg(CC1100_MCSM2, 0x07);	// Configure RX_TIME (until end of packet)
		spiStrobe(CC1100_SRX);
		wait_ticks(CC1100_IDLE_TO_RX_TIME);
		radioState = RADIO_RX;
	}
}




