/* Copyright (C) 2005, 2006, 2007, 2008 by Thomas Hillebrandt and Heiko Will

This file is part of the Micro-mesh SensorWeb Firmware.

Micro-Mesh is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3, or (at your option)
any later version.

Micro-Mesh is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Micro-Mesh; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

/* As a special exception, if you include this header file into source
   files compiled by GCC, this header file does not by itself cause
   the resulting executable to be covered by the GNU General Public
   License.  This exception does not however invalidate any other
   reasons why the executable file might be covered by the GNU General
   Public License.  */

#ifndef DRIVER_TIMER_H_
#define DRIVER_TIMER_H_

#include "type.h"

// --------------------------------------------------------
// Global Vars
// --------------------------------------------------------

// This four Vars are needed for the preemptive task-switcher
// which is und development. Soon they should be moved in a struct.
extern volatile int sheduler_sr;		// Stores the r2 register from the interrupted usertask
extern volatile int sheduler_back;		// Indicates if a usertask was interruptet as least once
extern volatile int sheduler_main_sr;	// Stores the r2 register of the inaktive main task 
extern volatile int sheduler_main;		// Stores the programmcounter of the inactive main task




// --------------------------------------------------------
// Public functions of the Sheduler
// TODO: Move them to a seperate Module
// --------------------------------------------------------

//-------------------------------------------------------------------------------------------------------
//  void activate_sheduler(void)
//
//  DESCRIPTION:
//		Activates the sheduler. This has to be called by the kernel, *directly before* calling 
//		a function of the userspace.
//
//---------------------------------------------------------------------------------------------------
extern void activate_sheduler(void);


//-------------------------------------------------------------------------------------------------------
//  void disable_sheduler(void)
//
//  DESCRIPTION:
//		Halts the sheduler. This has to be called by the kernel *directly after* returning to his 
//		own kontext.
//		TODO: move into the sheduler and make it private!
//
//---------------------------------------------------------------------------------------------------
extern void disable_sheduler(void);

//-------------------------------------------------------------------------------------------------------
//  void switch_userapp(void)
//
//  DESCRIPTION:
//		Switches from the Kernel Kontext to an interrupted userspace function. This has to be called
//		by the kernel after finishing a mainloop cycle *and if* there is a pending userspace function. 
//
//---------------------------------------------------------------------------------------------------
extern void switch_userapp(void);

//-------------------------------------------------------------------------------------------------------
//  extern void sleep(unsigned int secs)
//
//  DESCRIPTION:
//		Does a task-switch and switches back after secs seconds
//
//	ARGUMENTS:
//      unsigned int secs
//          Number of sceonds to sleep.
//
//---------------------------------------------------------------------------------------------------
extern void sleep(unsigned int secs);

//-------------------------------------------------------------------------------------------------------
//  void start_user_thread(callback_t thread)
//
//  DESCRIPTION:
//		Starts a function in the userspace context and guarantees a clean end of it.
//		This is called by the kernel to run a function in the userspace kontext.
//		TODO: This should handle most of the lowlevel switching. Should be the only public function.
//
//  ARGUMENTS:
//      callback_t thread
//          The function to start in userspace.
//
//---------------------------------------------------------------------------------------------------
extern void start_user_thread(callback_t thread);

// --------------------------------------------------------
// Timer-Api internal Functions (normaly used by timer.c)
// --------------------------------------------------------

//-------------------------------------------------------------------------------------------------------
//  void driver_timer_load(callback_t timer_proc)
//
//  DESCRIPTION:
//		Initializes the timer Hardware on the msp430. This is called frome the init_timer() function
//		and mustn't called by hand in the micro-mesh firmware.
//
//  ARGUMENTS:
//      callback_t timer_proc
//          The callback for any proced timer. Called from an interrupt.
//
//---------------------------------------------------------------------------------------------------
extern void driver_timer_load(callback_t timer_proc);

//-------------------------------------------------------------------------------------------------------
//  void set_next_timer_value(unsigned int ticks)
//
//  DESCRIPTION:
//		Sets the *absolute* ticks *offset* of the next timer to proc.
//		On 16 Bit Machines, like the msp430 this is only called when the next timer
//		is in the same 16Bit offset range like the timer register. 
//
//  ARGUMENTS:
//      unsigned int num
//          The absolute tickcount to wait for.
//
//---------------------------------------------------------------------------------------------------
extern void set_next_timer_value(unsigned int ticks);

//-------------------------------------------------------------------------------------------------------
//  void disable_timer(void)
//
//  DESCRIPTION:
//		Disables the timer logic. All tick counters won't stop. But no timer will proc.
//
//---------------------------------------------------------------------------------------------------
extern void disable_timer(void);

//-------------------------------------------------------------------------------------------------------
//  void _wait(unsigned int num)
//
//  DESCRIPTION:
//		A wait function which could be invoked by low-level drivers which could'nt use the full timer Api
//		TODO: Reconstruct the TimerApi. So that even device driver could use it.
//
//  ARGUMENTS:
//      unsigned int num
//          The total number of ticks to wait.
//
//---------------------------------------------------------------------------------------------------
extern void _wait(unsigned int num);

//-------------------------------------------------------------------------------------------------------
//  uint32_t *TICKS(void)
//
//  DESCRIPTION:
//		The Number of ticks since system reboot. 
//		TODO: This function is not overflow safe!
//		The Length of a tick is defined in cc1100_defines.h of the Hardware branch. Use the
//		TIMER_TICK_USEC_RES Define to normalize.
//		Only used by low-level hardware drivers.
//		TODO: Reconstruct the TimerApi. So that even device driver could use it.
//
//  RETURN VALUE:
//      uint32_t
//          A pointer to the Number of ticks since system reboot.
//			Returns a pointer because the msp430 gcc has a bug in returning 32Bitters
//
//---------------------------------------------------------------------------------------------------
extern uint32_t *TICKS(void);

//-------------------------------------------------------------------------------------------------------
//  unsigned int get_timer_count(void)
//
//  DESCRIPTION:
//		Returns the value of the timer register. On 32Bit Microcontroller this should be the same as 
//		*TICKS(), but on 16 Bitters its only the 16Bit offset of *TICKS().
//
//  RETURN VALUE:
//      int
//          The actual value of the timer register.
//---------------------------------------------------------------------------------------------------
extern unsigned int get_timer_count(void);

// --------------------------------------------------------
// Private functions of the Sheduler
// TODO: Move them to a seperate Module
// --------------------------------------------------------
extern void continue_thread(void);

#endif /*DRIVER_TIMER_H_*/
