/* Copyright (C) 2005, 2006, 2007, 2008 by Thomas Hillebrandt and Heiko Will

This file is part of the Micro-mesh SensorWeb Firmware.

Micro-Mesh is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3, or (at your option)
any later version.

Micro-Mesh is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Micro-Mesh; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include "smallib.h"
#include "driver_system.h"
#include "driver_controller.h"
#include "type.h"

void IRQ_Routine (void)   __attribute__ ((interrupt("IRQ")));
void FIQ_Routine (void)   __attribute__ ((interrupt("FIQ")));
void SWI_Routine (void)   __attribute__ ((interrupt("SWI")));
void UNDEF_Routine (void) __attribute__ ((interrupt("UNDEF")));

void IRQ_Routine (void) {
	sl_printf("irq");	
}

void FIQ_Routine (void)  {
	sl_printf("fiq");	
}


void SWI_Routine (void)  {
	sl_printf("swi");	
}


void UNDEF_Routine (void) {
	sl_printf("error");	
	while(1){};
}

#define PLOCK 0x400

static void feed(void)
{
	PLL0FEED = 0xAA;
	PLL0FEED = 0x55;
}

void init_clks(void)
{
	
	// 				Setting the Phased Lock Loop (PLL)
	//               ----------------------------------
	//
	// Olimex LPC-P2148 has a 12.0000 mhz crystal
	//
	// We'd like the LPC2148 to run at 60 mhz (has to be an even multiple of crystal)
	// 
	// According to the Philips LPC2148 manual:   M = cclk / Fosc	where:	M    = PLL multiplier (bits 0-4 of PLLCFG)
	//																		cclk = 60000000 hz
	//																		Fosc = 12000000 hz
	//
	// Solving:	M = 60000000 / 12000000 = 5           
	//
	//			Note: M - 1 must be entered into bits 0-4 of PLLCFG (assign 4 to these bits)
	//
	//
	// The Current Controlled Oscilator (CCO) must operate in the range 156 mhz to 320 mhz
	//
	// According to the Philips LPC2148 manual:	Fcco = cclk * 2 * P    where:	Fcco = CCO frequency 
	//																			cclk = 60000000 hz
	//																			P = PLL divisor (bits 5-6 of PLLCFG)
	//
	// Solving:	Fcco = 60000000 * 2 * P
	//			P = 2  (trial value)
	//			Fcco = 60000000 * 2 * 2
	//			Fcc0 = 240000000 hz    (good choice for P since it's within the 156 mhz to 320 mhz range)
	//
	// From Table 22 (page 34) of Philips LPC2148 manual    P = 2, PLLCFG bits 5-6 = 1  (assign 1 to these bits)
	//
	// Finally:      PLLCFG = 0  01  00100  =  0x24
	//
	// Final note: to load PLLCFG register, we must use the 0xAA followed 0x55 write sequence to the PLLFEED register
	//             this is done in the short function feed() below
	//
   
	// Setting Multiplier and Divider values
  	PLL0CFG = 0x24;
  	feed();
  
	// Enabling the PLL */	
	PLL0CON = 0x1;
	feed();
  
	// Wait for the PLL to lock to set frequency
	while(!(PLL0STAT & PLOCK)) ;
  
	// Connect the PLL as the clock source
	PLL0CON = 0x3;
	feed();
  
	// Enabling MAM and setting number of clocks used for Flash memory fetch
	MAMTIM = 0x3;
	MAMCR = 0x2;
  
	// Setting peripheral Clock (pclk) to 1/2 System Clock (cclk)
	VPBDIV = 0x2;
}

void init_ports(void)  
{
	PCON|= BIT5; //Disable BrownOut Detection Reset
	
	SCS |= BIT0 + BIT1 ; // Set IO Ports to fast switching mode
	FIO0DIR |= BIT0+BIT2+BIT4+BIT6+BIT7+BIT8+BIT11+BIT13+BIT17+BIT19+BIT20+BIT21+BIT22+BIT25;
	FIO1DIR |= BIT16+BIT17+BIT18+BIT19+BIT20+BIT21+BIT23+BIT24+BIT25+BIT26+BIT27+BIT28+BIT29+BIT30+BIT31;
	
	FIO0SET = BIT4+BIT7+BIT2+BIT11+BIT13+BIT17+BIT19+BIT20+BIT21+BIT22; 
	FIO1SET = BIT20+BIT23+BIT24+BIT26+BIT28+BIT29;
}

