#include <msp430x16x.h>
#include <signal.h>
#include "driver_adc.h"
#include "driver_timer.h"
#include "../funk.h"
#include "smallib.h"
#include "../timer.h"
#include "../boot.h"

#define FILTER_SIZE 29

static void init_epma(void);

// Endpoint moving average (EPMA)
static int weight_sum;
static int weights[FILTER_SIZE];

// Send counter
static uint8_t count;

// Storage of calculated filter values for x, y and z
static uint16_t xyz[3];
static uint32_t xyz_avg[3];

// Ringbuffer for x, y and z values
static uint8_t filter_size;
static uint8_t filter_in, filter_out;
static uint16_t filter[3][FILTER_SIZE];

void init_ADC(void) {
	ADC12CTL0 = SHT0_15 + SHT1_15 + MSC; 		// Number of ADC12CLK cycles in sampling period (=1024)
	ADC12CTL1 = SHP + CONSEQ_3 + CSTARTADD_0; 	// Use sampling timer, Repeat-sequence-of-channels, start at ADC12MEM0
	//ADC12CTL1 |= ADC12DIV_4;					// Devide ADC12CLK by 5
	ADC12MCTL0 = (INCH_0 + SREF_0);				// Input channel: A0 (A0 = P6.0)
	ADC12MCTL1 = (INCH_0 + SREF_0);				// Input channel: A0 (A0 = P6.0)
	ADC12MCTL2 = (INCH_0 + SREF_0);				// Input channel: A0 (A0 = P6.0)
	ADC12MCTL3 = (INCH_1 + SREF_0);				// Input channel: A1 (A1 = P6.1)
	ADC12MCTL4 = (INCH_1 + SREF_0);				// Input channel: A1 (A1 = P6.1)
	ADC12MCTL5 = (INCH_1 + SREF_0);				// Input channel: A1 (A1 = P6.1)
	ADC12MCTL6 = (INCH_2 + SREF_0);				// Input channel: A2 (A2 = P6.2)
	ADC12MCTL7 = (INCH_2 + SREF_0);				// Input channel: A2 (A2 = P6.2)
	ADC12MCTL8 = (INCH_2 + SREF_0);				// Input channel: A2 (A2 = P6.2)
	ADC12MCTL9 = (INCH_10 + SREF_0) + EOS;		// Input channel: Temperature sensor, End of conversion sequence

	filter_size = filter_in = filter_out = 0;
	count = 0;
	xyz[0] = 0;
	xyz[1] = 0;
	xyz[2] = 0;
	init_epma();
}

int enable_ADC(void) {
	init_ADC();
	ADC12CTL0 |= ADC12ON;			// Turn on ADC12
	_wait(50);						// Delay for reference start-up
	ADC12IE = 0x0200;
	ADC12CTL0 |= ENC + ADC12SC;		// Enable & start conversion
	return 0;
}

int disable_ADC(void) {
	ADC12CTL0 &= ~(ENC);			// Stop conversion
	while (ADC12CTL1 & ADC12BUSY);	// Wait until ADC12 is not busy
	ADC12CTL0 &= ~(ADC12ON);		// Turn off ADC12
	return 0;
}

int set_3D_sensitivity(int sensitivity) {
	switch (sensitivity)
	{
		case ACCELEROMETER_SENSITIVITY_1_5G:
			{P5OUT &= ~0x30; break;}
		case ACCELEROMETER_SENSITIVITY_2G:
			{P5OUT &= ~0x20; P5OUT |= 0x10; break;}
		case ACCELEROMETER_SENSITIVITY_4G:
			{P5OUT &= ~0x10; P5OUT |= 0x20; break;}
		case ACCELEROMETER_SENSITIVITY_6G:
			{P5OUT |=  0x30; break;}
	}
	return 0;
}

void init_epma(void)
{
	int tmp;
	uint8_t i;
	weight_sum = 0;
	tmp = 2 * FILTER_SIZE - 1;
	for (i = 0; i < FILTER_SIZE; i++)
	{
		weights[i] = tmp;
		weight_sum += tmp;
		tmp -= 3;
	}
}

void filter_add0(uint16_t *x, uint16_t *y, uint16_t *z)
{
	filter[0][filter_in] = *x;
	filter[1][filter_in] = *y;
	filter[2][filter_in] = *z;
	filter_in = (filter_in + 1) % FILTER_SIZE;
	filter_size++;
}

int filter_add(uint16_t *x, uint16_t *y, uint16_t *z)
{
	if (filter_size < FILTER_SIZE - 1)
	{
		filter_add0(x, y, z);
		return -10000;
	}
	else
	{
		uint8_t fin = filter_in;
		uint8_t i, j = FILTER_SIZE - 1;
		filter_add0(x, y, z);
		xyz_avg[0] = 0;
		xyz_avg[1] = 0;
		xyz_avg[2] = 0;
		for (i = filter_out; i != fin; i = ((i + 1) % FILTER_SIZE))
		{
			xyz_avg[0] += (weights[j] * (uint32_t)filter[0][i]);
			xyz_avg[1] += (weights[j] * (uint32_t)filter[1][i]);
			xyz_avg[2] += (weights[j] * (uint32_t)filter[2][i]);
			j--;
		}
		// Add for last value in ring buffer
		xyz_avg[0] += (weights[j] * (uint32_t)filter[0][fin]);
		xyz_avg[1] += (weights[j] * (uint32_t)filter[1][fin]);
		xyz_avg[2] += (weights[j] * (uint32_t)filter[2][fin]);

		*x = (uint16_t)(xyz_avg[0] / weight_sum);
		*y = (uint16_t)(xyz_avg[1] / weight_sum);
		*z = (uint16_t)(xyz_avg[2] / weight_sum);
		filter_out = (filter_out + 1) % FILTER_SIZE;
		filter_size--;
		return 0;
	}
}

static volatile int breakHysteresis = 0;
static void breakLEDoff(void)
{
	breakHysteresis = 0;
	P4OUT |= 0x02;	//BREAK LED OFF
}
static uint16_t pktc = 0;


void send_xyz()
{
	static uint8_t buffer[18];
	pktc += 1;
	buffer[0] = 17;					// Length
	buffer[1] = ADDR_BROADCAST;		// Broadcast
	((uint16_t*)buffer)[1] = xyz[0];
	((uint16_t*)buffer)[2] = xyz[1];
	((uint16_t*)buffer)[3] = xyz[2];
	((uint16_t*)buffer)[4] = pktc;

	((uint16_t*)buffer)[5] = (ADC12MEM0 + ADC12MEM1 + ADC12MEM2)/3; // RAW X
	((uint16_t*)buffer)[6] = (ADC12MEM3 + ADC12MEM4 + ADC12MEM5)/3; // RAW Y
	((uint16_t*)buffer)[7] = (ADC12MEM6 + ADC12MEM7 + ADC12MEM8)/3; // RAW Z

//BRAKE LED Routine----------------------------------------------------------
/*
	int breakVal = xyz[0];
	breakVal -= 2048;
	breakVal = sl_abs(breakVal);
	if (breakVal > XACCTHRESHOLD && breakHysteresis == 0)
	{
		P4OUT &= ~0x02;  //BREAK LED ON
		breakHysteresis  = register_timer(breakLEDoff, BREAKTIME*4);
	}
*/
//BRAKE LED Routine end-------------------------------------------------------

	sendPacket(buffer, 18);
}

interrupt(ADC_VECTOR) adc_isr(void)
{
	//P4OUT ^= 0x02; // Sample rate: 476 Hz
	uint16_t x, y, z, t;
	x = (ADC12MEM0);
	y = (ADC12MEM3);
	z = (ADC12MEM6);
	t = filter_add(&x, &y, &z);

	x = (ADC12MEM1);
	y = (ADC12MEM4);
	z = (ADC12MEM7);
	t = filter_add(&x, &y, &z);

	x = (ADC12MEM2);
	y = (ADC12MEM5);
	z = (ADC12MEM8);
	t = filter_add(&x, &y, &z);

	if (t != -10000)
	{
		xyz[0] += x;
		xyz[1] += y;
		xyz[2] += z;
		count++;
		if (count == 5)
		{
			xyz[0] /= 5;
			xyz[1] /= 5;
			xyz[2] /= 5;
			send_xyz();
			count = 0;
			xyz[0] = 0;
			xyz[1] = 0;
			xyz[2] = 0;
		}
	}
	ADC12IFG = 0; // Clear IFG
}
