/* Copyright (C) 2005, 2006, 2007, 2008 by Thomas Hillebrandt and Heiko Will

This file is part of the Micro-mesh SensorWeb Firmware.

Micro-Mesh is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3, or (at your option)
any later version.

Micro-Mesh is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Micro-Mesh; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include <msp430x16x.h> 
#include <signal.h>
#include "type.h"
#include "cc1100_defines.h"
#include "driver_cc1100.h"
#include "driver_system.h"
#include "spi0.h"

static callback_t _paket_cb; 
static callback_t _cs_cb;

//-------------------------------------------------------------------------------------------------------
// Public CC1100 communication functions (SPI)
//-------------------------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------------------------
//  void spiInitTrx(void)
//
//  DESCRIPTION:
//		This function puts the cc1100 into spi mode. You have to call this bevore every spi transaction.
//  
//-------------------------------------------------------------------------------------------------------
void drivercc1100_spiinittrx(void)
{
	// Switch to GDO mode
	P3SEL &= ~0x04;
	P3DIR &= ~0x04;
	// CS to low
	CC1100_CS_LOW;
	/* Wait for SO to go low (voltage regulator
	 * has stabilized and the crystal is running) */
	long c = 0;
	while (CC1100_GDO1)
	{
		if (c++ == 10000000) 
			alarm();
	}
	/* Switch to SPI mode */
	P3SEL |= 0x04;
}

uint8_t drivercc1100_trxspi(uint8_t data)
{
	/* Ensure TX Buf is empty */
	long c = 0;
	IFG1 &= ~UTXIFG0;
	IFG1 &= ~URXIFG0;
	TXBUF0 = data;
	while(!(IFG1 & UTXIFG0))
	{
		if (c++ == 1000000) 
			alarm();
	}
	/* Wait for Byte received */
	c = 0;
	while(!(IFG1 & URXIFG0))
	{
		if (c++ == 1000000) 
			alarm();	
	}
	return RXBUF0;
}

void drivercc1100_spiwriteburstreg(uint8_t addr, unsigned char *buffer, uint8_t count)
{
	uint8_t i;
	long c;
	drivercc1100_spiinittrx();
	drivercc1100_trxspi(addr | CC1100_WRITE_BURST);
	for (i = 0; i < count; i++)
	{
		c = 0;
		IFG1 &= ~UTXIFG0;
		IFG1 &= ~URXIFG0;
		TXBUF0 = buffer[i];
		/* Wait for TX to finish */
		while(!(IFG1 & UTXIFG0))
		{
			if (c++ == 1000000) 
				alarm();
		}
	}
	/* Wait for Byte received */
	c = 0;
	while(!(IFG1 & URXIFG0))
	{
		if (c++ == 1000000) 
			alarm();
	}
	CC1100_CS_HIGH;
}

void drivercc1100_spireadburstreg(uint8_t addr, char *buffer, uint8_t count)
{
	uint8_t i;
	drivercc1100_spiinittrx();
	drivercc1100_trxspi(addr | CC1100_READ_BURST);
	for (i = 0; i < count; i++)
	{
		long c = 0;
		IFG1 &= ~UTXIFG0;
		IFG1 &= ~URXIFG0;
		TXBUF0 = NOBYTE;
		while(!(IFG1 & UTXIFG0))
		{
			if (c++ == 1000000) 
				alarm();
		}
		/* Wait for Byte received */
		c = 0;
		while(!(IFG1 & URXIFG0))
		{
			if (c++ == 1000000) 
				alarm();
		}
		buffer[i] = RXBUF0;
	}
	CC1100_CS_HIGH;
}

void drivercc1100_load(callback_t cs_cb,callback_t paket_cb)
{
	_paket_cb = paket_cb;
	_cs_cb = cs_cb;
	spi0_init(0);	
}

void drivercc1100_aftersend(void)
{
    CLEAR(P2IFG, 0x01);	
	SET(P2IE, 0x01); /* Enable interrupts on port 2 pin 0 */
	CLEAR(P4OUT, 0x08); /* Turn off Sending Led*/
}

void drivercc1100_initinterrupts(void)
{
	_DINT(); /* Disable all interrupts */
	P2SEL = 0x00; /* must be <> 1 to use interrupts */
	SET(P2IES, 0x01); /* Enables external interrupt on low edge (for GDO2) */
	SET(P2IE, 0x01); /* Enable interrupt */
    CLEAR(P2IFG, 0x01); /* Clears the interrupt flag */
	CLEAR(P2IE, 0x02); /* Disable interrupt for GDO0 */
	CLEAR(P2IFG, 0x02); /* Clear IFG for GDO0 */
	_EINT(); /* Enable all interrupts */
}

void drivercc1100_beforesend(void)
{
	/* Turn on Led while sending paket for debug reasons */
	SET(P4OUT, 0x08);
	/* Disable interrupts on port 2 pin 0 */
	CLEAR(P2IE, 0x01);
}

void drivercc1100_enablecsdetect(void)
{
	CLEAR(P2IFG, 0x02);	/* Clear IFG for GDO0 */
	SET(P2IE, 0x02); /* Enable interrupt for GDO0 */ 
}

void drivercc1100_disablecsdetect(void)
{
	CLEAR(P2IE, 0x02);	/* Disable interrupt for GDO0 */
	CLEAR(P2IFG, 0x02);	/* Clear IFG for GDO0 */
}

/*
 * Private functions
 */


/*
 * CC1100 receive interrupt
 */
interrupt (PORT2_VECTOR) wor_isr(void){
	if (system_state.POWERDOWN) SPI_INIT; /* Initialize SPI after wakeup */
	/* Check IFG */
	if ((P2IFG & 0x01) != 0) {
		CLEAR(P2IFG, 0x01);
		_paket_cb();
	}
	else if ((P2IFG & 0x02) != 0) {
		_cs_cb();
		CLEAR(P2IE, 0x02);	// Disable interrupt for GDO0
		CLEAR(P2IFG, 0x02);	// Clear IFG for GDO0
	} else {
		/* Should not occur - only Port 2 Pin 0 interrupts are enabled */
		CLEAR(P2IFG, 0xFF);	/* Clear all flags */
	}
	if (system_state.POWERDOWN != 0) END_LPM3;
}
