/* Copyright (C) 2005, 2006, 2007, 2008 by Thomas Hillebrandt and Heiko Will

This file is part of the Micro-mesh SensorWeb Firmware.

Micro-Mesh is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3, or (at your option)
any later version.

Micro-Mesh is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Micro-Mesh; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include "funk.h"
#include "type.h"
#include "timer.h"
#include "defaultSettings.h"
#include "routing.h"
#include "smallib.h"
#include "helpers.h"
#include "driver_timer.h"
#include "driver_system.h"

#define MAX_TABLE_LENGTH		(50)	// Maximum length of the routing table
#define MAX_SEQ_BUFFER_LENGTH	(20)	// Maximum length of the sequence number buffer
#define MAX_TX_BUFFER_SIZE		(05)	// Maximum length of the txBuffer
#define MAX_RX_BUFFER_SIZE		(15)	// Maximum length of the rxBuffer

#define TTL						(10)	// Hop limit
#define SEND_RETRY_LIMIT		(03)	// Retry limit for route discovery
#define MAX_ROUTE_LENGTH		(48)	// Maximum route length (not changeable)
#define MAX_PAYLOAD_LENGTH		(50)	// Maximum length of data (not changeable)

static app_callback_t appCallback;

//------------------------- The routing table ------------------------------

typedef struct
{
	uint8_t destination;	// Destination address.
	uint8_t router;		// Node to use as router.
	uint8_t hops;			// Hopcount (to destination).
	uint32_t lastUsage;	// Last usage time
} routeEntry;

static routeEntry routes[MAX_TABLE_LENGTH];

//------------------------- Routing layer flags ----------------------------

typedef struct
{
	unsigned C_FLAG	: 1;	// Should C-Flag be set?
	unsigned unused	: 7;	// Unused
} layer1_flag_t;

static layer1_flag_t routingFlags;

//--------------------- The sequence number buffer -------------------------

typedef struct
{
	uint32_t sequenceNumber;	// Sequence number
	uint8_t seqClearId;		// Sequence number clear id, 0 if normal sequence number
} bcSeqEntry; /* Broadcast sequence number entry */

typedef struct
{
	uint32_t sequenceNumber;	// Sequence number
	uint32_t timestamp;		// Creation time of this entry
} ucSeqEntry; /* Unicast sequence number entry */

static volatile uint16_t seqNum = 0;									// Sequence number counter
static volatile bcSeqEntry seqNumBuffer[MAX_SEQ_BUFFER_LENGTH]; 	// Sequence number buffer (broadcast)
static volatile uint8_t seqNumBufferPos = 0;							// Next pos to enter a new value into ::seqNumBuffer

static volatile ucSeqEntry unicastSeqBuffer[MAX_SEQ_BUFFER_LENGTH]; // Sequence number buffer (unicast)
static volatile uint8_t unicastSeqBufferPos = 0;						// Next pos to enter a new value into ::unicastSeqBuffer

//--------------------- Packet types of routing ----------------------------

#define P_RREQ    (1)	// Protocol identifier for route request
#define P_RREP    (2)	// Protocol identifier for route reply
#define P_RERR    (3)	// Protocol identifier for route error
#define P_CMD   (100)	// Protocol identifier for network command (flooded)

//----------------------------------------------------------------------
// Layer 2 protocol (routing layer "Route-Request and Route-Reply"):
//
// ------------------------------------------
// |                 |                      |
// |  Length + Dummy |        Route         |
// |                 |                      |
// ------------------------------------------
//     2 byte               <= 48 bytes
//
//----------------------------------------------------------------------
typedef struct  __attribute__ ((packed)) p_layer2_rr
{
	uint8_t length;
	uint8_t dummy;
	uint8_t data[MAX_ROUTE_LENGTH];
} packet_layer2_rr_t;

#define DEST_NOT_REACHABLE  	(1)	// Route error type: destination not reachable
#define DATA_DELIVERY_FAILURE 	(2)	// Route error type: data delivery failure

//----------------------------------------------------------------------
// Layer 2 protocol (routing layer "Route-Error"):
//
// ------------------------------------------------------
// |                       |        |                   |
// |  OriginalDestination  |  Type  |     Padding       |
// |                       |        |                   |
// ------------------------------------------------------
//          1 byte           1 byte        48 bytes
//
//----------------------------------------------------------------------
typedef struct __attribute__ ((packed)) p_layer2_re
{
	uint8_t originalDestination;
	uint8_t type;
	uint8_t padding[MAX_PAYLOAD_LENGTH-2];
} packet_layer2_re_t;

//----------------------------------------------------------------------
// Layer 2 protocol (routing layer "Command"):
//
// ------------------------------------------------------
// |                       |        |                   |
// |      CmdSource        |  CmdId |     CmdData       |
// |                       |        |                   |
// ------------------------------------------------------
//          1 byte           1 byte        <= 48 bytes
//
//----------------------------------------------------------------------
typedef struct  __attribute__ ((packed)) p_layer2_cmd
{
	uint8_t cmdSource;
	uint8_t cmdId;
	uint8_t cmdData[MAX_PAYLOAD_LENGTH-2];
} packet_layer2_cmd_t;

//--------------------- TX buffer of routing --------------------------------

typedef struct
{
	packet_layer1_t packet;	// Payload (user packet)
	uint32_t timestamp;		// Time of last route request for this packet
	uint8_t retryCount;		// Route request counter for this packet
} packetRecord;

static packetRecord dsrTxBuffer[MAX_TX_BUFFER_SIZE];	// TX buffer

//--------------------- RX buffer of routing --------------------------------

static volatile packet_layer0_t dsrRxBuffer[MAX_RX_BUFFER_SIZE];	// RX buffer (FIFO)

static volatile uint8_t rxBufferPutIdx;	// Next free position in ::dsrRxBuffer (if any)
static volatile uint8_t rxBufferGetIdx;	// Next packet in ::dsrRxBuffer to read out (if any)
static volatile uint8_t rxBufferSize;		// The current size of the ::dsrRxBuffer

static packetRecord* lastPacketSent;	// Pointer to last sent packet in buffer
								// (for update of timestamps in route requests)

static uint32_t timeout;					// Timeout for new RREQ
static uint32_t timeoutOneHop;			// Timeout for new RREQ with TTL = 1

//---------------------------------------------------------------------------
//					F U N C T I O N S   of the routing layer
//---------------------------------------------------------------------------

void initRouting(app_callback_t callback)
{
	int i;
	appCallback = callback;
	routingFlags.C_FLAG = true;
	rxBufferPutIdx = rxBufferGetIdx = rxBufferSize = 0;
	
	uint32_t sendtime = (SEND_TIME + TIMEOUT_SEND) * BURST_COUNT;
	timeout = 4 * (sendtime + DIFS) * TTL;
	timeoutOneHop = 4 * (sendtime + DIFS);

	for (i = 0; i < MAX_TABLE_LENGTH; i++)
	{
		routes[i].destination = 0;
	}
	for (i = 0; i < MAX_TX_BUFFER_SIZE; i++)
	{
		dsrTxBuffer[i].timestamp = 0;
	}
	for (i = 0; i < MAX_SEQ_BUFFER_LENGTH; i++)
	{
		seqNumBuffer[i].sequenceNumber = 0;
		seqNumBuffer[i].seqClearId = 0;
		unicastSeqBuffer[i].sequenceNumber = 0;
	}
}

//---------------------------------------------------------------------------
// Routing table management
//---------------------------------------------------------------------------

uint8_t findDestinationIndex(uint8_t dest)
{
	int i = 0;
	while (i < MAX_TABLE_LENGTH && routes[i].destination != dest) i++;
	return i;
}

routeEntry* getRoute(uint8_t dest)
{
	uint8_t rTablePos = findDestinationIndex(dest);
	if (rTablePos == MAX_TABLE_LENGTH) return NULL;
	return &(routes[rTablePos]);
}

void addRouteEntry(uint8_t dest, uint8_t router, uint8_t hops)
{
	// For LRU search
	uint8_t i;
	uint32_t lru = (2 ^ (sizeof(uint32_t)*8)) - 1;
	// Is the destination already in the table
	uint8_t rTablePos = findDestinationIndex(dest);
	// Not found, then find free position
	if (rTablePos == MAX_TABLE_LENGTH) rTablePos = findDestinationIndex(0);
	// No free position, then overwrite entry least recently used
	if (rTablePos == MAX_TABLE_LENGTH)
	{
		rTablePos = 0;
		for (i = 0; i < MAX_TABLE_LENGTH; i++)
		{
			if (routes[i].lastUsage < lru)
			{
				lru = routes[i].lastUsage;
				rTablePos = i;
			}
		}
	}
	routes[rTablePos].destination = dest;
	routes[rTablePos].router = router;
	routes[rTablePos].hops = hops;
	routes[rTablePos].lastUsage = *TICKS();
}

void addRoute(uint8_t myUid, uint8_t length, uint8_t* list)
{
	uint8_t i = 0;
	while (i < length && list[i] != myUid) i++;
	if (i == length) return;

	uint8_t pos = i;
	int leftNeighbour = -1;
	int rightNeighbour = -1;
	if (pos > 0)
	{
		leftNeighbour = list[pos-1];
	}
	if (pos+1 != length)
	{
		rightNeighbour = list[pos+1];
	}

	i = 0;
	while (i < length)
	{
		uint8_t next = list[i];
		if (myUid != next)
		{
			int distance = pos - i;
			int router = leftNeighbour;
			if (distance < 0) router = rightNeighbour;
			addRouteEntry(next, (uint8_t)router, (uint8_t)sl_abs(distance));
		}
		i++;
	}
}

void removeRoute(uint8_t dest)
{
	uint8_t rTablePos = findDestinationIndex(dest);
	if (rTablePos == MAX_TABLE_LENGTH) return;
	routes[rTablePos].destination = 0;
}

void removeRouter(uint8_t router)
{
	uint8_t i;
	for (i = 0; i < MAX_TABLE_LENGTH; i++)
	{
		if (routes[i].router == router)
		{
			routes[i].destination = 0;
		}
	}
}

//---------------------------------------------------------------------------
// Sequence number buffer management
//---------------------------------------------------------------------------

void makeSeqNum(uint16_t seq, uint8_t id, uint32_t* ret)
{
	uint32_t new_seq_num = id;
    new_seq_num <<= 16;
    new_seq_num += seq;
	*ret = new_seq_num;
}

bool containsSeqNum(uint32_t* n, uint8_t c)
{
	int i;
	for (i = 0; i < MAX_SEQ_BUFFER_LENGTH; i++)
	{
		if (seqNumBuffer[i].sequenceNumber == *n && seqNumBuffer[i].seqClearId == c)
		{
			return true;
		}
	}
	return false;
}

void addSeqNum(uint32_t* n, uint8_t c)
{
	if (containsSeqNum(n, c)) return;
	seqNumBuffer[seqNumBufferPos].sequenceNumber = *n;
	seqNumBuffer[seqNumBufferPos++].seqClearId = c;
	if (seqNumBufferPos == MAX_SEQ_BUFFER_LENGTH) seqNumBufferPos = 0;
}

bool containsUnicastSeqNum(uint32_t* n)
{
	uint8_t i;
	uint32_t now = *TICKS();
	for (i = 0; i < MAX_SEQ_BUFFER_LENGTH; i++)
	{
		if (unicastSeqBuffer[i].sequenceNumber == *n && (now - unicastSeqBuffer[i].timestamp) <= timeoutOneHop)
		{
			return true;
	    }
	}
	return false;
}

void addUnicastSeqNum(uint32_t* n)
{
	if (containsUnicastSeqNum(n)) return;
	unicastSeqBuffer[unicastSeqBufferPos].sequenceNumber = *n;
	unicastSeqBuffer[unicastSeqBufferPos++].timestamp = *TICKS();
	if (unicastSeqBufferPos == MAX_SEQ_BUFFER_LENGTH) unicastSeqBufferPos = 0;
}

void makeAndAddSeqNum(uint16_t n, uint8_t c)
{
	uint32_t tmp;
	makeSeqNum(n, uid, &tmp);
	addSeqNum(&tmp, c);
}

void makeAndAddUnicastSeqNum(uint16_t n)
{
	uint32_t tmp;
	makeSeqNum(n, uid, &tmp);
	addUnicastSeqNum(&tmp);
}

uint16_t getNextSeqNum(void)
{
	if (seqNum == 0xFFFF)
	{
		seqNum = 0; // => Sequence number always != 0
		routingFlags.C_FLAG = true; // Overflow detected, set C-Flag
	}
	seqNum++;		// don't write return seqNum++;
	return seqNum;	// seqNum doesn't get increased this way!
}

//---------------------------------------------------------------------------
// RX buffer management
//---------------------------------------------------------------------------

bool rxBufferAdd(packet_layer0_t* packet)
{
	uint8_t i;
	uint32_t tmp, mask;
	
	// Any space in rxBuffer left?
	if (rxBufferSize == MAX_RX_BUFFER_SIZE) return false;
	
	// Check sequence numbers by packet type
	packet_layer1_t* dsr = (packet_layer1_t*)(packet->data);
	makeSeqNum(dsr->identification, dsr->source, &tmp);
	
	/* Broadcast */
	if (packet->address == ADDR_BROADCAST)
	{
		if (!containsSeqNum(&tmp, dsr->seqClearId))
		{
			if (READ_CFLAG(dsr->flags))
			{
				// Delete all sequence numbers of this node
				mask = dsr->source;
				mask = (mask << 16) & 0xFFFF0000;
				for (i = 0; i < MAX_SEQ_BUFFER_LENGTH; i++)
				{
					if ((seqNumBuffer[i].sequenceNumber & mask) == mask)
					{
						seqNumBuffer[i].sequenceNumber = 0;
						seqNumBuffer[i].seqClearId = 0;
				    }
				}
			}
			addSeqNum(&tmp, dsr->seqClearId);
		}
		else
		{
			return false;
		}
	}
	/* Unicast */
	else
	{
		if (!containsUnicastSeqNum(&tmp))
		{
			addUnicastSeqNum(&tmp);
		}
		else
		{
			return false;
		}
	}
	
	rxBufferSize++;
	dsrRxBuffer[rxBufferPutIdx] = *packet;
	rxBufferPutIdx++;
	if (rxBufferPutIdx == MAX_RX_BUFFER_SIZE) rxBufferPutIdx = 0;
	return true;
}

void rxBufferRemove(void)
{
	if (rxBufferSize == 0) return;
	rxBufferGetIdx++;
	rxBufferSize--;
	if (rxBufferGetIdx == MAX_RX_BUFFER_SIZE) rxBufferGetIdx = 0;
}

//---------------------------------------------------------------------------
// Packet handling (sending and receiving)
//---------------------------------------------------------------------------

bool sendLayer1Data(uint8_t* data, uint8_t appProtocol, uint8_t dest)
{
	uint8_t i;
	uint8_t j;
	
	if (dest == uid) return false;
	if (appProtocol < P_DATA) return false;
	i = 0;
	while (i < MAX_TX_BUFFER_SIZE && dsrTxBuffer[i].timestamp != 0) i++;
	if (i == MAX_TX_BUFFER_SIZE) return false;
	dsrTxBuffer[i].retryCount = 0;
	dsrTxBuffer[i].timestamp = 1;
	dsrTxBuffer[i].packet.protocol = appProtocol;
	dsrTxBuffer[i].packet.destination = dest;
	dsrTxBuffer[i].packet.source = uid;
	dsrTxBuffer[i].packet.identification = getNextSeqNum();
	dsrTxBuffer[i].packet.flags = 0;
	dsrTxBuffer[i].packet.seqClearId = 0;
	dsrTxBuffer[i].packet.ttl = TTL;
	for (j = 0; j < MAX_PAYLOAD_LENGTH; j++)
	{
		dsrTxBuffer[i].packet.data[j] = data[j];
	}
	return true;
}

bool sendLayer1Command(uint8_t command, uint8_t* data)
{
	uint8_t j;
	packet_layer1_t packet;
	packet.protocol = P_CMD;
	packet.destination = 0;
	packet.source = uid;
	packet.identification = getNextSeqNum();
	packet.flags = 0;
	packet.seqClearId = 0;
	packet.ttl = TTL;
	
	if (routingFlags.C_FLAG)
	{
		packet.flags = WRITE_CFLAG(packet.flags);
		packet.seqClearId = irand(255) + 1; // 1..255
		routingFlags.C_FLAG = false;
	}
	
	// Build and store sequence number to avoid cycles
	makeAndAddSeqNum(packet.identification, packet.seqClearId);
	
	packet.data[0] = uid;
	packet.data[1] = command;
	for (j = 2; j < MAX_PAYLOAD_LENGTH; j++)
	{
		packet.data[j] = data[j-2];
	}
	return csend(ADDR_BROADCAST, L1_AAODV, (uint8_t*)&packet, sizeof(packet));
}

void sendLayer1RouteRequest(uint8_t dest)
{
	uint8_t i;
	packet_layer1_t packet;
	
	// Clear packet data to 0
	for (i = 0; i < sizeof(packet_layer1_t); i++)
	{
		((uint8_t*)&packet)[i] = 0;
	}
	
	packet.protocol = P_RREQ;
	packet.destination = dest;
	packet.source = uid;
	packet.identification = getNextSeqNum();
	packet.flags = 0;
	packet.seqClearId = 0;
	if (lastPacketSent->retryCount == 1)
	{
		packet.ttl = 1;
	}
	else
	{
		packet.ttl = TTL;
	}
	
	if (routingFlags.C_FLAG)
	{
		packet.flags = WRITE_CFLAG(packet.flags);
		packet.seqClearId = (uint8_t) (irand(255) + 1); // 1..255
		routingFlags.C_FLAG = false;
	}
	
	// Build and store sequence number to avoid cycles
	makeAndAddSeqNum(packet.identification, packet.seqClearId);
	
	packet.data[0] = 1;		// Set length of route to 1
	packet.data[1] = 0;		// Set dummy to 0
	packet.data[2] = uid;	// Put our node to list
	csend(ADDR_BROADCAST, L1_AAODV, (uint8_t*)&packet, sizeof(packet_layer1_t));
	if (lastPacketSent != NULL) lastPacketSent->timestamp = *TICKS();	// Update timestamp to now
}

bool sendLayer1RouteReply(packet_layer2_rr_t* rr, uint8_t dest, uint8_t phyDest)
{
	uint8_t i;
	packet_layer1_t packet;
	packet.protocol = P_RREP;
	packet.destination = dest;
	packet.source = uid;
	packet.identification = getNextSeqNum();
	packet.flags = 0;
	packet.seqClearId = 0;
	packet.ttl = TTL;
	packet_layer2_rr_t* a = (packet_layer2_rr_t*)(packet.data);
	a->length = rr->length;
	for (i = 0; i < rr->length; i++)
	{
		a->data[i] = rr->data[i];
	}
	
	// Build and store sequence number
    makeAndAddUnicastSeqNum(packet.identification);
	
	return csend(phyDest, L1_AAODV, (uint8_t*)&packet, sizeof(packet_layer1_t));
}

bool sendLayer1RouteError(uint8_t dest, uint8_t phyDest, uint8_t notReachableAddress, uint8_t type)
{
	packet_layer1_t packet;
	packet.protocol = P_RERR;
	packet.destination = dest;
	packet.source = uid;
	packet.identification = getNextSeqNum();
	packet.flags = 0;
	packet.seqClearId = 0;
	packet.ttl = TTL;
	packet_layer2_re_t* re = (packet_layer2_re_t*) (packet.data);
	re->originalDestination = notReachableAddress;
	re->type = type;
	
	// Build and store sequence number
    makeAndAddUnicastSeqNum(packet.identification);
	
	return csend(phyDest, L1_AAODV, (uint8_t*)&packet, sizeof(packet_layer1_t));
}

bool isPacketHandlingNeeded(void)
{
	int i;
	
	if (rxBufferSize > 0) return true;
	
	for (i = 0; i < MAX_TX_BUFFER_SIZE; i++)
	{
		if (dsrTxBuffer[i].timestamp != 0)
		{
			routeEntry* re = getRoute(dsrTxBuffer[i].packet.destination);
			if (re != NULL)
			{
				return true;
			}
			else
			{
				if (dsrTxBuffer[i].retryCount == 0) return true;
				if (dsrTxBuffer[i].retryCount == 1)
				{
					if (*TICKS() - dsrTxBuffer[i].timestamp > timeoutOneHop)
					{
						return true;
					}
				}
				if (*TICKS() - dsrTxBuffer[i].timestamp > timeout)
				{
					return true;
				}
			}
		}
	}

	return false;
}

void handleIncomingPackets(void)
{
	if (rxBufferSize <= 0) return;
	packet_layer0_t packet = dsrRxBuffer[rxBufferGetIdx];
	rxBufferRemove(); // Delete packet

	uint8_t protocol = R_FLAGS_PROTOCOL(packet.flags);

	if (protocol == L1_AAODV)
	{
		packet_layer1_t* dsr = (packet_layer1_t*)(packet.data);
		dsr->ttl--;

		if (dsr->ttl == 0 && dsr->destination != uid)
		{
			return;
		}

		if (packet.address == uid)
		{
			if (dsr->protocol >= P_DATA)
			{
				if (dsr->destination == uid)
				{
					if (appCallback != NULL) appCallback(dsr);
				}
				else
				{
					routeEntry* re = getRoute(dsr->destination);
					if (re != NULL)
					{
						bool sent = csend(re->router, L1_AAODV, (uint8_t*)dsr, sizeof(packet_layer1_t));
						if (!sent) 
						{
							sendLayer1RouteError(dsr->source, packet.phySrc, dsr->destination, DATA_DELIVERY_FAILURE);
						}
						else
						{
							re->lastUsage = *TICKS();
						}
					}
					else
					{
						re = getRoute(dsr->source);
						bool sent;
						if (re != NULL)
						{
							sent = sendLayer1RouteError(dsr->source, re->router, dsr->destination, DEST_NOT_REACHABLE);
						}
						else
						{
							sent = sendLayer1RouteError(dsr->source, packet.phySrc, dsr->destination, DEST_NOT_REACHABLE);
						}
						if (sent && re != NULL) re->lastUsage = *TICKS();
					}
				}
			}

			if (dsr->protocol == P_RERR)
			{
				packet_layer2_re_t* routeError = (packet_layer2_re_t*)(dsr->data);
				
				if (dsr->destination == uid)
				{
					if (routeError->type == DEST_NOT_REACHABLE)
					{
						removeRoute(routeError->originalDestination);
					}
					else if (routeError->type == DATA_DELIVERY_FAILURE)
					{
						removeRoute(routeError->originalDestination);
					}
				}
				else
				{
					// Remove route to destination in RERR packet
					removeRoute(routeError->originalDestination);
					// Forward packet
					routeEntry* re = getRoute(dsr->destination);
					if (re != NULL)
					{
						bool sent = csend(re->router, L1_AAODV, (uint8_t*)dsr, sizeof(packet_layer1_t));
						if (sent) 
						{
							re->lastUsage = *TICKS();
						}
					}
				}
			}

			if (dsr->protocol == P_RREP)
			{
				packet_layer2_rr_t* rr = (packet_layer2_rr_t*)(dsr->data);
				uint8_t* list = rr->data;
				addRoute(uid, rr->length, list);

				if (dsr->destination != uid)
				{
					uint8_t i = 0;
					while (i < rr->length && list[i] != uid) i++;
					bool sent = csend(list[i-1], L1_AAODV, (uint8_t*)dsr, sizeof(packet_layer1_t));
					if (!sent)
					{
						sendLayer1RouteError(dsr->source, packet.phySrc, dsr->destination, DEST_NOT_REACHABLE);
					}
				}
			}
		}

		if (packet.address == ADDR_BROADCAST)
		{
			if (dsr->protocol == P_RREQ)
			{
				packet_layer2_rr_t* rr = (packet_layer2_rr_t*)(dsr->data);
				uint8_t* list = rr->data;
				if (rr->length < MAX_ROUTE_LENGTH)
				{
					list[rr->length++] = uid;
					addRoute(uid, rr->length, list);	// add route with overhearing
				}
				else
				{
					// Distance between sender and receiver to long -> Do not answer.
					// => TTL <= MAX_ROUTE_LENGTH.
					return;
				}

				if (dsr->destination == uid)
				{
					bool sent = sendLayer1RouteReply(rr, dsr->source, list[rr->length-2]);
					if (!sent)
					{
						removeRouter(list[rr->length-2]); // remove all routes with list[rr->length-2] as router!
					}
				}
				else
				{
					csend(ADDR_BROADCAST, L1_AAODV, (uint8_t*)dsr, sizeof(packet_layer1_t));
				}
			}

			if (dsr->protocol == P_CMD)
			{
				// Global network command
				uint32_t tmp;
				
				// Flood network
				csend(ADDR_BROADCAST, L1_AAODV, (uint8_t*)dsr, sizeof(packet_layer1_t));

				// Execute command after broadcast
				packet_layer2_cmd_t* cmd = (packet_layer2_cmd_t*)(dsr->data);
				switch (cmd->cmdId)
				{
					case COMMAND_RESET:
						tmp = *TICKS() + 170000;	// ~20 sec
						while (*TICKS() < tmp) {
							Led1Toggle();		// wait and toggle LED to indicate reset
							wait_ticks(5460);
						}
						McuReset();
						break;
					case COMMAND_OUTPUT_POWER:
						setOutputPower(cmd->cmdData[0]);
						break;
					default:
						break;
				}
			}
		}
	}
}

void handleOutgoingPackets(void)
{
	uint8_t i;
	for (i = 0; i < MAX_TX_BUFFER_SIZE; i++)
	{
		if (dsrTxBuffer[i].timestamp != 0)
		{
			routeEntry* re = getRoute(dsrTxBuffer[i].packet.destination);
			if (re != NULL)
			{
				bool sent = csend(re->router, L1_AAODV, (uint8_t*)&(dsrTxBuffer[i].packet), sizeof(packet_layer1_t));
				if (sent)
				{
					makeAndAddUnicastSeqNum(dsrTxBuffer[i].packet.identification); // Store SEQ-Num
					dsrTxBuffer[i].timestamp = 0; // Delete packet out of buffer
					re->lastUsage = *TICKS();	  // Update route usage time
				}
				else
				{
					removeRoute(dsrTxBuffer[i].packet.destination); // Remove broken route
					dsrTxBuffer[i].retryCount = 0;	// For immediate route request
				}
				return;
			}
			else
			{
				if (dsrTxBuffer[i].retryCount == 0) 
				{
					// Packet has not been send yet, send Route-Request now
					dsrTxBuffer[i].retryCount++;
					lastPacketSent = &dsrTxBuffer[i];
					sendLayer1RouteRequest(dsrTxBuffer[i].packet.destination);
					return;
				}
				
				// Routing optimization (TTL=1)
				if (dsrTxBuffer[i].retryCount == 1)
				{
					if (*TICKS() - dsrTxBuffer[i].timestamp > timeoutOneHop)
					{
						dsrTxBuffer[i].retryCount++;
						lastPacketSent = &dsrTxBuffer[i];
						sendLayer1RouteRequest(dsrTxBuffer[i].packet.destination);
						return;
					}
				}

				if (*TICKS() - dsrTxBuffer[i].timestamp > timeout)
				{
					if (dsrTxBuffer[i].retryCount == SEND_RETRY_LIMIT)
					{
						// Delete all packets with current destination out of buffer;
						// TODO: Eventually send DEST_NOT_REACHABLE message to application
						// (error callback)
						uint8_t j;
						uint8_t deleteId = dsrTxBuffer[i].packet.destination;
						for (j = 0; j < MAX_TX_BUFFER_SIZE; j++)
						{
							if (dsrTxBuffer[j].timestamp != 0
									&& dsrTxBuffer[j].packet.destination == deleteId)
							{
								dsrTxBuffer[j].timestamp = 0;	// Delete packet out of buffer
							}
						}
						lastPacketSent = (packetRecord*) 0; // Set packet pointer to null
						return;
					}
	
					// Timeout is over, send new Route-Request
					dsrTxBuffer[i].retryCount++;
					lastPacketSent = &(dsrTxBuffer[i]);
					sendLayer1RouteRequest(dsrTxBuffer[i].packet.destination);
					return;
				}
			}
		}
	}
}

void handlePackets(void)
{
	handleIncomingPackets();
	handleOutgoingPackets();
}

void getNextRouteRequestTimeout(uint32_t* _timeout)
{
	uint8_t i;
	uint32_t callTime;
	*_timeout = 0;
	for (i = 0; i < MAX_TX_BUFFER_SIZE; i++)
	{
		if (dsrTxBuffer[i].timestamp != 0)
		{
			if (dsrTxBuffer[i].retryCount == 1)
			{
				callTime = dsrTxBuffer[i].timestamp + timeoutOneHop;
			}
			else
			{
				callTime = dsrTxBuffer[i].timestamp + timeout;
			}
			*_timeout = (*_timeout != 0) ? (callTime < *_timeout ? callTime : *_timeout) : callTime;
		}
	}
}

//---------------------------------------------------------------------------
// Debug stuff (packet and routing table printing etc.)
//---------------------------------------------------------------------------

void printRoutingTable(void)
{
	uint8_t i;
	for (i = 0; i < MAX_TABLE_LENGTH; i++)
	{
		if (routes[i].destination != 0)
		{
			sl_printf("  Destination: %u Router: %u Hops: %u LastUsage: %lu\n",
					routes[i].destination, routes[i].router, routes[i].hops, routes[i].lastUsage);
		}
	}
}

void printSeqNumBuffer(void)
{
	uint8_t i;
	sl_printf("<< Broadcast buffer >>\n");
	for (i = 0; i < MAX_SEQ_BUFFER_LENGTH; i++)
	{
		if (seqNumBuffer[i].sequenceNumber != 0)
		{
			sl_printf("SEQ = %lu ", seqNumBuffer[i].sequenceNumber);
			sl_printf("SEQ_CLEAR_ID = %u\n", seqNumBuffer[i].seqClearId);
		}
	}
	sl_printf("\n<< Unicast buffer >>\n");
	for (i = 0; i < MAX_SEQ_BUFFER_LENGTH; i++)
	{
		if (unicastSeqBuffer[i].sequenceNumber != 0)
		{
			sl_printf("SEQ = %lu ", unicastSeqBuffer[i].sequenceNumber);
			sl_printf("TIMESTAMP = %lu\n", unicastSeqBuffer[i].timestamp);
		}
	}
}
