/* Copyright (C) 2005, 2006, 2007, 2008 by Thomas Hillebrandt and Heiko Will

 This file is part of the Micro-mesh SensorWeb Firmware.

 Micro-Mesh is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 Micro-Mesh is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Micro-Mesh; see the file COPYING.  If not, write to
 the Free Software Foundation, 59 Temple Place - Suite 330,
 Boston, MA 02111-1307, USA.  */

#include "user_pakethandlers.h"
#include "type.h"
#include "net.h"
#include "smallib.h"
#include "command.h"
#include "helpers.h"
#include "display.h"
#include "driver_sht11.h"
#include "boot.h"
#include "config.h"
#include "rulesystem.h"

#ifdef INCLUDE_SHT11
void sht11(uint8_t source, uint8_t* data);
#endif
void echo(uint8_t source, uint8_t* data);
void print(uint8_t source, uint8_t* data);
void sht11in(uint8_t source, uint8_t* data);
void sensor_in(uint8_t source, uint8_t* data);
void rulesystem(uint8_t source, uint8_t* data);

int init_user_paket_handlers(void) {
	registerPacketHandler(127, print);
	registerPacketHandler(128, echo);
#ifdef INCLUDE_SHT11
	registerPacketHandler(129, sht11);
	registerPacketHandler(140, serial_request);
	registerPacketHandler(141, serial_response);
#endif
	registerPacketHandler(130, sht11in);
	registerPacketHandler(131, sensor_in);
	registerPacketHandler(150, rulesystem);
	return 0;
}

//Packet Handler
void echo(uint8_t source, uint8_t* data) {
	net_send(127, data[0], data+1);
	sl_printf("Chat [%u]: %s\n", source, data+1);
}

void print(uint8_t source, uint8_t* data) {
	data[51]=0;
	sl_printf("Got message [%u]: %s\n", source, data);
	execCommand('r');
}

#ifdef INCLUDE_SHT11
void sht11(uint8_t source, uint8_t* data) {
	uint8_t pos=0;
	uint8_t buffer[52];
	sht11_update_data();
	sl_memcpy(buffer+pos, &sht11_data.temp_val.f, sizeof(float));
	pos+=sizeof(float);
	int h = (int)sht11_data.humi_val.f;
	sl_memcpy(buffer+pos, &h, sizeof(uint16_t));
	net_send(130, source, buffer);
}
#endif

void sht11in(uint8_t source, uint8_t* data) {
	char cbuf[6];
	float temp=0;
	uint16_t hum;
	sl_memcpy(&temp, data, sizeof(float));
	sl_memcpy(&hum, data+sizeof(float), sizeof(uint16_t));
	float2string(cbuf, temp, 2, 3, 1);
	sl_printf("SensorData from Node %u: Temperature=%s°C, Humidity=%u%%\n",
			source, cbuf, hum);
}

void rulesystem(uint8_t source, uint8_t* data) {
	rule_protocoll_t *rp = (rule_protocoll_t*) data;
	unsigned char id;
	
	switch (rp->command) {
	case RULE_ADDRULE:
		id = add_rule(&rp->rule);
		sl_printf("Added rule %i: \n", id);
		printrule(&rp->rule);
		rp->command = RULE_ADDREPLY;
		rp->rule_id = id;
		net_send(150, source, (uint8_t*)rp);
		break;
	case RULE_ADDREPLY:
		sl_printf("Rule has been added with id %u: \n", rp->rule_id);
		break;
	case RULE_DELRULE:
		del_rule(rp->rule_id);
		sl_printf("Rule #%u has been deleted.\n", rp->rule_id);
		rp->command = RULE_DELREPLY;
		net_send(150, source, (uint8_t*)rp);
		break;
	case RULE_DELREPLY:
		sl_printf("Rule #%u has been deleted.\n", rp->rule_id);
		break;
	}
}

static const char *sensors[] = { "", "Temperature", "Humidity", "Battery" };

void sensor_in(uint8_t source, uint8_t* data) {
	static int xx=0;
	char cbuf[6];
	float fval = 0;
	uint16_t ival = 0;
	int pos=1;
	uint16_t sensor_type=0;
	uint8_t t= data[0];
#ifdef INCLUDE_SDCARD
#ifdef INCLUDE_FAT16
	if (log_file_open)
	{
		sl_sprintf(l_buf,"    <node id=\"%u\">\n",source);
		file_write(&log_file,l_buf,sl_strlen(l_buf));
	}
#endif
#endif
	while (t!=0) {
		if (t == STYPE_FLOAT) {
			sl_memcpy(&fval, data+pos, sizeof(float));
			pos+=sizeof(float);
			float2string(cbuf, fval, 2, 3, 1);
		}
		if (t == STYPE_UINT16) {
			sl_memcpy(&ival, data+pos, sizeof(uint16_t));
			pos+=sizeof(uint16_t);
		}
		sl_memcpy(&sensor_type, data+pos, sizeof(uint16_t));
		pos+=sizeof(uint16_t);
		if (t == STYPE_FLOAT) {
#ifdef INCLUDE_SDCARD
#ifdef INCLUDE_FAT16
			if (log_file_open)
			{
				sl_sprintf(l_buf,"        <sensor id=\"%u\">\n            <value float=\"%s\" />\n        </sensor>\n",sensor_type,cbuf);
				file_write(&log_file,l_buf,sl_strlen(l_buf));
			}
#endif
#endif
			sl_printf("SensorData from Node %u: %s = %s\n", source,
					sensors[sensor_type], cbuf);
			char buffer[16];
			sl_sprintf(buffer, "Paket in: %u\n", xx++);glcd_write(buffer);
		}
		if (t == STYPE_UINT16) {
#ifdef INCLUDE_SDCARD
#ifdef INCLUDE_FAT16
			if (log_file_open)
			{
				sl_sprintf(l_buf,"        <sensor id=\"%u\">\n            <value int=\"%u\" />\n        </sensor>\n",sensor_type,ival);
				file_write(&log_file,l_buf,sl_strlen(l_buf));
			}
#endif
#endif
			sl_printf("SensorData from Node %u: %s = %u\n", source,
					sensors[sensor_type], ival);
		}
		t=data[pos++];
	}
#ifdef INCLUDE_SDCARD
#ifdef INCLUDE_FAT16
	if (log_file_open)
	{
		sl_sprintf(l_buf,"    </node>\n");
		file_write(&log_file,l_buf,sl_strlen(l_buf));
		uart_init(115200,0);
	}
#endif
#endif
}
