package carrera;

import carrera.data.CarreraGame;
import carrera.data.CarreraGameEvent;
import carrera.data.CarreraGameListener;
import carrera.data.CarreraGames;
import carrera.data.GameEndTask;
import carrera.data.GameModeSwitcher;
import carrera.data.MP3Player;
import carrera.data.SerialGameCommands;
import carrera.data.plot.PlotData;
import carrera.data.plot.PlotDriver;
import carrera.serial.SerialConnectionManager;
import java.awt.Component;
import java.util.Vector;
import javax.swing.UIManager;

/**
 *
 * @author hillebra
 */
public class Main {

    private MainView gui;
    private OSSupport osSupport;
    
    private MP3Player player;
    
    /** Spieldaten */
    private int gameMode;
    private CarreraGames games;
    private CarreraGame currentGame;
    private Vector<CarreraGameListener> listeners;
    
    /** Tasks */
    private GameModeSwitcher gms;
    private GameEndTask get;
    
    /** Anzeige der Beschleunigung im Plot */
    private PlotData plotData;
    
    /** Plot driver - only for testing */
    private PlotDriver plotDriver;

    /** Daten aus dem Einstellungsdialog */
    private int maxSpeed = 14;
    private int minSpeed = 5;
    private int breakTime = 600;
    private int xThreshold = 700;
    private boolean showBT = true;
    
    /** Status des automatischen Autos */
    private int vehicleStatus = -1;
    
    /** Laden und Speichern von Spieldaten */
    public static final String GAME_FILE = "carrera_games.ser";
    public static final String PROGRAM_SAVE_PATH = "carrera";
    public static final String PROGRAM_NAME = "Carrera Game VCU";
    public static final String PROGRAM_VERSION = "1.2";
    public static final String PROGRAM_COPYRIGHT = "Copyright 2008-2009 FUB/Inf/AG Tech.";

    public static String lastUserName = "";
    
    public Main() {
        
        osSupport = new OSSupport();
        player = new MP3Player();
                
        try {
	    // Set System L&F
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch (Exception e) {
            System.err.println(e.getLocalizedMessage());
        }

        plotData = new PlotData();
        listeners = new Vector<CarreraGameListener>();
        
        if ((games = openCarreraGames(new java.io.File(
                osSupport.getUserHomeDirectory() + GAME_FILE))) == null) {
            games = new CarreraGames();
        }
        
        get = new GameEndTask(this);
        gms = new GameModeSwitcher(this, 15);
        gms.start();
        
        gui = new MainView(this);
        show(gui);
        
        // Only for testing
        //plotDriver = new PlotDriver(plotData);
        //plotDriver.start();
    }
    
    /**
     * Application entry point.
     * 
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        new Main();
    }
    
    public MP3Player getPlayer() {
        return player;
    }

    public int getBreakTime() {
        return breakTime;
    }

    public int getMaxSpeed() {
        return maxSpeed;
    }

    public int getMinSpeed() {
        return minSpeed;
    }

    public int getXThreshold() {
        return xThreshold;
    }

    public boolean isShowBT() {
        return showBT;
    }

    public int getVehicleStatus() {
        return vehicleStatus;
    }

    public void setVehicleStatus(int vehicleStatus) {
        this.vehicleStatus = vehicleStatus;
    }
    
    public void setProperties(int speedMax, int speedMin, int timeBreak, int xThreshold, boolean showBT) {
        this.maxSpeed = speedMax;
        this.minSpeed = speedMin;
        this.breakTime = timeBreak;
        this.xThreshold = xThreshold;
        this.showBT = showBT;
    }

    public PlotData getPlotData() {
        return plotData;
    }
    
    private int fail = 0;
    private int p_count, p_lastcount = 0;
    
    public void parseAndAddXYData(String s) {
        String[] strs = s.split(",");
        if (strs.length == 7) {
            try {
                p_count = Integer.parseInt(strs[3]);
                if (p_lastcount == 0) {
                    p_lastcount = p_count-1;
                }
                
                fail(p_count - p_lastcount - 1);
                p_lastcount = p_count;
                plotData.addAccData(Integer.parseInt(strs[0]),
                        Integer.parseInt(strs[1]), Integer.parseInt(strs[2]),
                        Integer.parseInt(strs[4]), Integer.parseInt(strs[5]),
                        Integer.parseInt(strs[6]));
            } catch (NumberFormatException e) {}
        } else {
            System.out.println("Cannot parse string: " + s);
        }
    }
    
    private void fail(int f) {
        if (f == 0) return;
        fail += f;
        System.out.println("LF: "+f+", GF: "+ fail);
    }
    
    /*************** Game Logic ****************/
    
    public void addGameListener(CarreraGameListener l) {
        if (!listeners.contains(l)) listeners.add(l);
    }
    
    public void removeGameListener(CarreraGameListener l) {
        listeners.remove(l);
    }
    
    public void notifyGameListeners(CarreraGameEvent event) {
        for (int i = 0; i < listeners.size(); i++) {
            listeners.elementAt(i).carreraGameChanged(event);
        }
    }
    
    public void initNewGame(String player1, String player2) {
        if (currentGame != null && !currentGame.isOver()) return;
        get.stop();
        gms.stop();
        currentGame = games.createGame(this, player2, player1);
        lastUserName = player1;
        setGameMode(CarreraGames.MODE_DISPLAY_CURRENT_GAME);
        player.play3();
        notifyGameListeners(new CarreraGameEvent(currentGame,
                CarreraGameEvent.EVENT_NEW_GAME));
    }
    
    public void startGame() {
        if (currentGame == null) return;
        vehicleStatus = -1;
        currentGame.start();
        notifyGameListeners(new CarreraGameEvent(currentGame,
                        CarreraGameEvent.EVENT_GAME_START));
    }
    
    public void endGame(boolean success) {
        if (currentGame == null) return;
        SerialConnectionManager.transmitData(gui.getSerialConnectionID(),
            SerialGameCommands.GAME_CMD_STOPP);
        currentGame.end(success);
        games.storeGame(currentGame);
        if (success) {
            get.setTime(10);
        } else {
            get.setTime(5);
        }
        get.start();
        notifyGameListeners(new CarreraGameEvent(currentGame,
                        CarreraGameEvent.EVENT_GAME_END));
        doGameSave(); // for backup reason
    }
    
    public void closeCurrentGame() {
        if (currentGame != null) {
            if (!currentGame.isOver()) endGame(false);
            setGameMode(CarreraGames.MODE_DISPLAY_BOOTSCREEN);
            notifyGameListeners(new CarreraGameEvent(currentGame,
                    CarreraGameEvent.EVENT_GAME_CLOSE));
            currentGame = null;
        }
    }
    
    public void notifyNewRound(int driverId) {
        if (currentGame != null) {
            currentGame.decreaseLapCount(driverId);
            notifyGameListeners(new CarreraGameEvent(currentGame,
                        CarreraGameEvent.EVENT_NEW_ROUND));
        }
    }
    
    public void transmitSerial(String s) {
        SerialConnectionManager.transmitData(gui.getSerialConnectionID(), s);
    }
    
    public CarreraGames getCarreraGames() {
        return games;
    }
    
    public CarreraGame getCurrentGame() {
        return currentGame;
    }
    
    public int getGameMode() {
        return gameMode;
    }
    
    public void setGameMode(int gameMode) {
        this.gameMode = gameMode;
        notifyGameListeners(new CarreraGameEvent(games,
                        CarreraGameEvent.EVENT_GAME_MODE_CHANGE));
    }
    
    public void enableGameModeSwitcher() {
        gms.start();
    }
    
    public void stopTasks() {
        get.stop();
        gms.stop();
    }

    /*************** Program Logic *************/
    
    public void show(final Component c) {
        java.awt.EventQueue.invokeLater(new Runnable() {
            public void run() {
                c.setVisible(true);
            }
        });
    }
    
    /**
     * Do things on application exit.
     */
    public void exit() {
        gms.stop();
        get.stop();
        if (plotDriver != null) plotDriver.finish();
        SerialConnectionManager.closeAllConnections();
        doGameSave();
        System.exit(0);
    }
    
    private void doGameSave() {
        if (osSupport.createUserHomeDirectory()) {
            saveCarreraGames(games, new java.io.File(
                osSupport.getUserHomeDirectory() + GAME_FILE));
        }
    }
    
    private CarreraGames openCarreraGames(java.io.File file) {
        if (file == null) return null;
        
        java.io.InputStream fis = null;
        java.io.ObjectInputStream ois = null;
        
        try {
            fis = new java.io.FileInputStream(file);
            ois = new java.io.ObjectInputStream(fis);
            CarreraGames gameData = (CarreraGames) ois.readObject();
            return gameData;
        } catch (Exception e) {
            return null;
        } finally {
            try { ois.close(); } catch (Exception e) {}
        }
    }
 
    private boolean saveCarreraGames(CarreraGames games, java.io.File file) {
        if (file == null) return false;
        
        /** Jetzt das Speichern mit ObjectOutputStream durchführen */
        java.io.ObjectOutputStream oos = null;
        java.io.OutputStream fos = null;
        
        try {
            fos = new java.io.FileOutputStream(file);
            oos = new java.io.ObjectOutputStream(fos);
            oos.writeObject(games);
            return true;
        } catch (java.io.IOException e) {
            return false;
        } finally { try { oos.close(); } catch (Exception e) {} }
    }
    
}
