package carrera.serial;

import gnu.io.CommPortIdentifier;
import gnu.io.NoSuchPortException;
import gnu.io.PortInUseException;
import gnu.io.SerialPort;
import gnu.io.SerialPortEvent;
import gnu.io.SerialPortEventListener;
import gnu.io.UnsupportedCommOperationException;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.util.TooManyListenersException;
import java.util.Vector;

/**
 * Stellt eine serielle Verbindung her.
 * 
 * @version 1.0, 2008-01-14
 * @author  Thomas Hillebrandt
 * @since   1.0
 */
public class SerialConnection implements Runnable, SerialPortEventListener {

    private SerialPort serialPort;
    private InputStream inputStream;
    private BufferedReader bufferedReader;
    private OutputStream outputStream;
    private SerialConnectionProperties scp;
    private Vector<SerialConnectionDataListener> listeners;
    
    private Thread t;
    
    /**
     * Erzeugt ein neues {@code SerialConnection}-Objekt.
     * 
     * @param scp Verbindungseinstellungen.
     */
    public SerialConnection(SerialConnectionProperties scp) {
        this.scp = scp;
        listeners = new Vector<SerialConnectionDataListener>();
    }
    
    /**
     * Registriert einen SerialConnectionDataListener.
     * 
     * @param l Der SerialConnectionDataListener.
     */
    public void addSerialConnectionDataListener(SerialConnectionDataListener l) {
        if (!listeners.contains(l)) listeners.add(l);
    }
    
    /**
     * Deregistriert einen SerialConnectionDataListener.
     * 
     * @param l Der SerialConnectionDataListener.
     */
    public void removeSerialConnectionDataListener(SerialConnectionDataListener l) {
        listeners.remove(l);
    }
    
    private void notifySerialConnectionDataEvent(SerialConnectionDataEvent event) {
        for (int i = 0; i < listeners.size(); i++) {
            listeners.elementAt(i).serialConnectionDataEvent(event);
        }
    }
    
    /**
     * Gibt einen Verbindungsnamen zur�ck.
     * 
     * @return Verbindungsname.
     */
    public String getName() {
        return scp.getPort();
    }
    
    /**
     * Stellt die Verbindung her.
     * 
     * @throws SerialConnectionException Falls ein Fehler beim Verbindungsaufbau
     *                                   auftritt.
     */
    public void connect() throws SerialConnectionException {
        try {
            CommPortIdentifier portId = CommPortIdentifier.getPortIdentifier(scp.getPort());
            serialPort = (SerialPort) portId.open(this.getClass().getName(), 2000);
            inputStream = serialPort.getInputStream();
            bufferedReader = new BufferedReader(new InputStreamReader(inputStream), 13);
            outputStream = serialPort.getOutputStream();
            serialPort.addEventListener(this);
            serialPort.notifyOnDataAvailable(true);
            serialPort.setSerialPortParams(scp.getBaudrate(),
                    scp.getDatabits(),
                    scp.getStopbits(),
                    scp.getParity());
            t = new Thread(this);
            t.start();
        } catch (NoSuchPortException e) {
            throw new SerialConnectionException(
                    SerialConnectionException.NO_SUCH_PORT);
        } catch (PortInUseException e) {
            throw new SerialConnectionException(
                    SerialConnectionException.PORT_IN_USE);
        } catch (IOException e) {
            throw new SerialConnectionException(
                    SerialConnectionException.IO_ERROR);
        } catch (TooManyListenersException e) {
            // Nichts machen...
        } catch (UnsupportedCommOperationException e) {
            throw new SerialConnectionException(
                    SerialConnectionException.UNSUPPORTED_COMM_OPERATION);
        }
    }
    
    /**
     * Trennt die Verbindung.
     */
    public void disconnect() {
        try {
            for (int i = 0; i < listeners.size(); i++) {
                listeners.elementAt(i).serialConnectionClosing();
            }
            bufferedReader.close();
            inputStream.close();
            outputStream.close();
        } catch (IOException e) {
            // Nichts machen...
        } finally {
            serialPort.close();
            serialPort = null;
            for (int i = 0; i < listeners.size(); i++) {
                listeners.elementAt(i).serialConnectionClosed();
            }
        }
    }
    
    /**
     * Schickt eine Zeichenkette über die serielle Schnittstelle.
     * 
     * @param s Eine Zeichenkette.
     */
    public void write(String s) {
        write(s.getBytes());
    }
    
    /**
     * Schickt die Daten �ber die serielle Schnittstelle.
     * 
     * @param data Die Daten.
     */
    public void write(byte[] data) {
        try {
            outputStream.write(data);
            outputStream.flush();
        } catch(IOException e) {}
    }

    public void run() {
        String s;
        while (true) {
            try {
                s = bufferedReader.readLine();
                notifySerialConnectionDataEvent(
                            new SerialConnectionDataEvent(this, s));
                Thread.yield();
            } catch (Exception e) {}
        }
    }
    
    @Override
    public void serialEvent(SerialPortEvent event) {
        switch (event.getEventType()) {
            case SerialPortEvent.BI:
            case SerialPortEvent.OE:
            case SerialPortEvent.FE:
            case SerialPortEvent.PE:
            case SerialPortEvent.CD:
            case SerialPortEvent.CTS:
            case SerialPortEvent.DSR:
            case SerialPortEvent.RI:
            case SerialPortEvent.OUTPUT_BUFFER_EMPTY:
                break;
            case SerialPortEvent.DATA_AVAILABLE:
                break;
        }
    }
   
}
