package carrera.serial;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

/**
 * Verwaltet alle seriellen Verbindungen.
 * 
 * @version 1.0, 2008-01-14
 * @author  Thomas Hillebrandt
 * @since   1.0
 */
public class SerialConnectionManager {

    private static int serialIDCount = 0;
    
    /** Tabelle mit allen offenen seriellen Verbindungen */
    private static Hashtable<Integer, SerialConnection> serialConnectionTable =
            new Hashtable<Integer, SerialConnection>();
    
    /** Alle SerialConnectionEventListener */
    private static Vector<SerialConnectionEventListener> listeners =
            new Vector<SerialConnectionEventListener>();
    
    /**
     * Registriert einen SerialConnectionEventListener.
     * 
     * @param l Der SerialConnectionEventListener.
     */
    public static void addSerialConnectionEventListener(SerialConnectionEventListener l) {
        if (!listeners.contains(l)) listeners.add(l);
    }
    
    /**
     * Deregistriert einen SerialConnectionEventListener.
     * 
     * @param l Der SerialConnectionEventListener.
     */
    public static void removeSerialConnectionEventListener(SerialConnectionEventListener l) {
        listeners.remove(l);
    }
    
    private static void notifySerialConnectionEvent(SerialConnectionEvent event) {
        for (int i = 0; i < listeners.size(); i++) {
            listeners.elementAt(i).serialConnectionEvent(event);
        }
    }
    
    /**
     * �ffnet den Dialog zum Erstellen einer neuen seriellen Verbindung.
     * 
     * @param source Die Quelle, die die neue serielle Verbindung erstellt.
     */
    public static void openConnectionDialog(SerialConnectionDataListener source) {
        notifySerialConnectionEvent(new SerialConnectionEvent(source,
                SerialConnectionEvent.SERIAL_CONNECTION_CREATE_REQUEST));
    }
    
    /**
     * �ffnet eine serielle Verbindung und speichert sie in der Tabelle.
     * 
     * @param source Die Quelle, die die neue serielle Verbindung erstellt.
     * @param scp Verbindungseinstellungen.
     * 
     * @return Die Verbindungskennung f�r die serielle Schnittstelle.
     * @throws SerialConnectionException Falls ein Fehler bei der Verbindung
     *                                   auftritt.
     */
    public static int openConnection(SerialConnectionDataListener source,
            SerialConnectionProperties scp) throws SerialConnectionException {
        if (source == null) {
            throw new SerialConnectionException(
                    SerialConnectionException.NO_DATA_LISTENER);
        }
        SerialConnection connection = new SerialConnection(scp);
        connection.addSerialConnectionDataListener(source);
        connection.connect();
        serialIDCount++;
        serialConnectionTable.put(serialIDCount, connection);
        source.serialConnectionEstablished(serialIDCount);
        return serialIDCount;
    }
    
    /**
     * Registriert einen SerialConnectionDataListener.
     * 
     * @param connectionID Die Verbindungskennung der Verbindung.
     * @param l Der SerialConnectionDataListener.
     */
    public static void addSerialConnectionDataListener(int connectionID,
            SerialConnectionDataListener l) {
        SerialConnection connection = serialConnectionTable.get(connectionID);
        if (connection != null) {
            connection.addSerialConnectionDataListener(l);
        }
    }
    
    /**
     * Deregistriert einen SerialConnectionDataListener.
     * 
     * @param connectionID Die Verbindungskennung der Verbindung.
     * @param l Der SerialConnectionDataListener.
     */
    public static void removeSerialConnectionDataListener(int connectionID,
            SerialConnectionDataListener l) {
        SerialConnection connection = serialConnectionTable.get(connectionID);
        if (connection != null) {
            connection.removeSerialConnectionDataListener(l);
        }
    }
    
    /**
     * Übermittelt eine Zeichenkette über die serielle Verbindung.
     * 
     * @param connectionID Die Verbindungskennung der Verbindung.
     * @param data Eine Zeichenkette, die übermittelt werden soll.
     */
    public static void transmitData(int connectionID, String s) {
        SerialConnection connection = serialConnectionTable.get(connectionID);
        if (connection != null) {
            connection.write(s);
        }
    }
    
    /**
     * �bermittelt die Daten �ber die serielle Verbindung.
     * 
     * @param connectionID Die Verbindungskennung der Verbindung.
     * @param data Die Daten, die �bermittelt werden sollen.
     */
    public static void transmitData(int connectionID, byte[] data) {
        SerialConnection connection = serialConnectionTable.get(connectionID);
        if (connection != null) {
            connection.write(data);
        }
    }
    
    /**
     * Schlie�t die Verbindung.
     * 
     * @param connectionID Die Verbindungskennung der Verbindung, die
     *                     geschlossen werden soll.
     */
    public static void closeConnection(int connectionID) {
        SerialConnection connection = serialConnectionTable.remove(connectionID);
        if (connection != null) {
            connection.disconnect();
        }
    }
    
    /**
     * Schlie�t alle offenen Verbindungen.
     */
    public static void closeAllConnections() {
        Iterator<Integer> iter = serialConnectionTable.keySet().iterator();
        while (iter.hasNext()) {
            SerialConnection connection = serialConnectionTable.get(iter.next());
            if (connection != null) {
                connection.disconnect();
            }
        }
        serialConnectionTable.clear();
        serialIDCount = 0;
    }
    
}
